# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""API related to node creation in topology"""

import logging

# pylint: disable=unused-import
# pylint: disable=cyclic-import
from nest import topology
from nest import engine
from nest.topology_map import TopologyMap
import nest.config as config
from nest.network_utilities import ipv6_dad_check
import nest.global_variables as g_var
from nest.input_validator import input_validator
from .address import Address
from .id_generator import IdGen

logger = logging.getLogger(__name__)


class Node:
    """
    Abstraction for a network namespace.

    Attributes
    ----------
    name: str
        User given name for the node.
    """

    @input_validator
    def __init__(self, name: str):
        """
        Create a node with given `name`.

        An unique `id` is assigned to this node which is used by
        `engine` module to create the network namespace.
        This ensures that there is no naming conflict between any two
        nodes.

        Parameters
        ----------
        name: str
            The name of the node to be created
        """
        if name == "":
            raise ValueError("Node name can't be an empty string")
        if config.get_value("assign_random_names") is False and len(name) > 3:
            # We chose 3 because: 'ifb-ns1-ns2-20' is a potential IFB interface name
            # and it's already 14 character long. Note that here node names
            # are 'ns1' and 'ns2'. The `ip` utility won't accept interface names
            # longer than 15 characters
            logger.warning(
                "%s is longer than 3 characters. It's safer to use "
                "node names with atmost 3 characters with the current config.",
                name,
            )

        self._name = name
        self._id = IdGen.get_id(name)
        self._interfaces = []
        # Global variable disables when any new node is created
        # to ensure DAD check (if applicable)
        g_var.IS_DAD_CHECKED = False

        engine.create_ns(self.id)
        engine.set_interface_mode(self.id, "lo", "up")
        TopologyMap.add_namespace(self.id, self.name)
        TopologyMap.add_host(self)

    def __enter__(self):
        """
        Enter the context of this `Node`.
        For eg., all commands in this context will only be able to see
        interfaces inside this `Node`
        """
        # Go to network namespace corresponding to `Node`
        engine.set_ns(self.id)

    def __exit__(self, *args):
        """
        Exit the context of this `Node`
        """
        # Switch back to default network namespace
        engine.set_ns(None)

    @input_validator
    def add_route(
        self,
        dest_addr: Address,
        via_interface: "topology.Interface",
        next_hop_addr: Address = None,
    ):
        """
        Add a route to the routing table of `Node`.

        Parameters
        ----------
        dest_addr: Address/str
            Destination IP address of node to route to. 'DEFAULT' is
            for all addresses
        via_interface: Interface
            `Interface` in `Node` to route via
        next_hop_addr: Address/str, optional
            IP address of next hop Node (or router), by default ''
        """
        if next_hop_addr is None:
            next_hop_addr = via_interface.pair.address

        dest_addr_str = ""
        if dest_addr.is_subnet():
            dest_addr_str = dest_addr.get_addr()
        else:
            dest_addr_str = dest_addr.get_addr(with_subnet=False)

        engine.add_route(
            self.id,
            dest_addr_str,
            next_hop_addr.get_addr(with_subnet=False),
            via_interface.id,
        )

    def get_interface(self, node, connection_number=1):
        """
        Get the interface in `self` connected to `node`.
        By default, this returns a veth interface.

        `connection_number` is an optional argument used when there
        multiple connections between `self` and `node`.

        Parameters
        ----------
        node: Node
            The other end point of the `veth` interface
        connection_number: int
            If there are multiple connections, then this argument
            uniquely identifies the connection

        Returns
        -------
        Interface
            Return the interface between `self` and `node`. Use
            `connection_number` to uniquly identify an interface if
            there are multiple connections.
            If no interface is found, then return None.
        """
        if connection_number <= 0:
            raise ValueError("connection_number should be greater than 0")

        for interface in self.interfaces:
            if hasattr(interface.pair, "node"):  # True if interface is a `veth`
                pair_node = interface.pair.node
                if node == pair_node:
                    connection_number -= 1

                if connection_number == 0:
                    return interface

        return None

    @input_validator
    def add_route_mpls_push(
        self, dest_addr: Address, next_hop_addr: Address, label: int
    ):
        """
        Add a route to the routing table of `Node`.

        Parameters
        ----------
        dest_addr: Address/str
            Destination ip address of node to route to. 'DEFAULT' is
            for all addresses
        next_hop_addr: Address/str, optional
            IP address of next hop Node (or router), by default ''
        label: the MPLS label pushed onto the packet
        """
        engine.add_mpls_route_push(
            self.id,
            dest_addr.get_addr(),
            next_hop_addr.get_addr(with_subnet=False),
            label,
        )

    @input_validator
    def add_route_mpls_switch(
        self, incoming_label: int, next_hop_addr: Address, outgoing_label: int
    ):
        """

        Parameters
        ----------
        incoming_label: The label that the packet carries when entering the interface
        next_hop_addr: IP address of the next hop
        outgoing_label: The label that is pushed onto the packet

        """
        engine.add_mpls_route_switch(
            self.id,
            incoming_label,
            next_hop_addr.get_addr(with_subnet=False),
            outgoing_label,
        )

    @input_validator
    def add_route_mpls_pop(self, incoming_label: int, next_hop_addr: Address):
        """

        Parameters
        ----------
        incoming_label: THe label that the packet carries when entering the interface
        next_hop_addr: IP address of the next hop

        """
        engine.add_mpls_route_pop(
            self.id, incoming_label, next_hop_addr.get_addr(with_subnet=False)
        )

    def _add_interface(self, interface):
        """
        Add `interface` to `Node`

        Parameters
        ----------
        interface: Interface
            `Interface` to be added to `Node`
        """
        self._interfaces.append(interface)
        interface.node_id = self.id
        engine.add_int_to_ns(self.id, interface.id)

    def configure_tcp_param(self, param, value):
        """
        Configure TCP parameters of `Node` available at
        /proc/sys/net/ipv4/tcp_*.

        Example: 'window_scaling', 'wmem', 'ecn', etc.

        If TCP Parameter `param` is valid, then new `value` is set
        for this `param`.

        Parameters
        ----------
        param: str
            TCP parameter to be configured
        value: str
            New value of TCP parameter `param`
        """
        engine.configure_kernel_param(self.id, "net.ipv4.tcp_", param, value)

    def configure_udp_param(self, param, value):
        """
        Configure UDP parameters of `Node` available at
        /proc/sys/net/ipv4/udp_*.

        Example: 'early_demux', 'l3mdev_accept', 'rmem_min', 'wmem_min'

        If UDP Parameter `param` is valid, then new `value` is set
        for this `param`.

        Parameters
        ----------
        param: str
            TCP parameter to be configured
        value: str
            New value of TCP parameter `param`
        """
        engine.configure_kernel_param(self.id, "net.ipv4.udp_", param, value)

    def read_tcp_param(self, param):
        """
        Read TCP parameters of `Node` available at
        `/proc/sys/net/ipv4/tcp_*`.

        Example: 'window_scaling', 'wmem', 'ecn', etc.

        Parameters
        ----------
        param: str
            TCP parameter to be read

        Returns
        -------
        str
            If TCP Parameter `param` is valid, then corresponding value
            is returned.
        """
        return engine.read_kernel_param(self.id, "net.ipv4.tcp_", param)

    def read_udp_param(self, param):
        """
        Read UDP parameters of `Node`available at
        `/proc/sys/net/ipv4/udp_*`.

        Example: 'early_demux', 'l3mdev_accept', 'rmem_min', 'wmem_min'

        Parameters
        ----------
        param: str
            UDP parameter to be read

        Returns
        -------
        str
            If UDP Parameter `param` is valid, then corresponding value
            is returned.
        """
        return engine.read_kernel_param(self.id, "net.ipv4.udp_", param)

    @ipv6_dad_check
    @input_validator
    def ping(
        self, destination_address: Address, packets: int = 5, verbose: bool = True
    ):
        """
        Ping from current `Node` to destination address
        if there is a route.

        Parameters
        ----------
        destination_address: Address/str
            IP address to ping to
        packets: int
            Number of ping packets sent
        verbose: bool
            If `True`, output ping packet details

        Returns
        -------
        bool
            `True` if `Node` can successfully ping `destination_address`.
            Else `False`.
        """
        if isinstance(destination_address, str):
            destination_address = Address(destination_address)

        if verbose:
            print()
            print(
                f"=== PING from {self.name} to "
                f"{destination_address.get_addr(with_subnet=False)} ==="
            )
            print()

        status = engine.ping(
            self.id,
            destination_address.get_addr(with_subnet=False),
            packets,
            destination_address.is_ipv6(),
            live_output=verbose,
        )

        return status

    def enable_ip_forwarding(self, ipv4=True, ipv6=True):
        """
        Enable IP forwarding in `Node`.

        After this method runs, the `Node` can be used as a router.
        """
        if not ipv4 and not ipv6:
            raise Exception(
                "IP Forwarding cannot be false for both IPv4 and IPv6 addresses"
            )

        engine.en_ip_forwarding(self.id, ipv4, ipv6)
        TopologyMap.add_router(self)

    def disable_ip_dad(self):
        """
        Disables Duplicate addresses Detection (DAD) for all interfaces of `Node`.
        """
        for i in range(len(self._interfaces)):
            interface_name = self._interfaces[i]
            interface_name.disable_ip_dad()

    @property
    def id(self):
        """
        str: Value used by `engine` to create the emulated `Node` entity
        """
        return self._id

    @property
    def interfaces(self):
        """list(Interface): List of interfaces in this node"""
        return self._interfaces

    @property
    def name(self):
        """str: User given name for `Node`"""
        return self._name

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.name!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Runs ss command and parses socket stats from
its output
"""

import os
import re
from functools import partial
from nest.experiment.interrupts import handle_keyboard_interrupt
from ..results import SsResults
from .runnerbase import Runner
from ...engine.iterators import run_ss


class SsRunner(Runner):
    """
    Runs ss command and stores and parses the output

    Attributes
    ----------
    iterator : str
        absolute path of the ss iterator script
    param_list: list(str)
        list of parameters to be parsed
    ns_id : str
        network namespace to run ss from
    destination_ip : str
        IP address of the destination namespace
    start_time : num
        time at which ss is to be run
    run_time : num
        total time to run ss for
    """

    iterator = os.path.realpath(os.path.dirname(__file__)) + "/iterators/ss.sh"
    param_list = [
        "cwnd",
        "rwnd",
        "rtt",
        "ssthresh",
        "rto",
        "delivery_rate",
        "pacing_rate",
    ]

    # pylint: disable=too-many-arguments
    def __init__(
        self, ns_id, destination_ip, start_time, run_time, dst_ns, ss_filter=""
    ):
        """
        Constructor to initialize ss runner

        Parameters
        ----------
        ns_id : str
            network namespace to run ss from
        destination_ip : str
            IP address of the destination namespace
        start_time : num
            time at which ss is to be run
        run_time : num
            total time to run ss for
        dst_ns : str
            destination network namespace of ss
        ss_filter : str
            to filter output from specific connections.
        """
        self.filter = ss_filter
        super().__init__(ns_id, start_time, run_time, destination_ip, dst_ns)

    def run(self):
        """
        Runs the ss iterator
        """

        super().run(
            partial(
                run_ss,
                self.ns_id,
                SsRunner.iterator,
                self.destination_address.get_addr(with_subnet=False),
                self.run_time,
                f'"{self.filter}"',
                self.start_time,
                self.destination_address.is_ipv6(),
            ),
            error_string_prefix="Collecting socket stats",
        )

    # pylint: disable=too-many-locals
    @handle_keyboard_interrupt
    def parse(self):
        """
        parses the required data from `self.out`
        """
        stats_dict_list = {}

        self.out.seek(0)  # rewind to start of the temp file

        # See `iterators/ss.h` for output format
        raw_stats = self.out.read().decode().split("---")
        destination_ip = self.destination_address.get_addr(with_subnet=False)

        for raw_stat in raw_stats[:-1]:

            # Pattern to capture port numbers of flows to `destination ip`
            if self.destination_address.is_ipv6():
                port_pattern = re.escape(destination_ip) + r"]:(?P<port>\d+)"
            else:
                port_pattern = re.escape(destination_ip) + r":(?P<port>\d+)"
            port_list = [
                port.group("port") for port in re.finditer(port_pattern, raw_stat)
            ]
            timestamp_pattern = r"timestamp:(?P<timestamp>\d+\.\d+)"
            timestamp = re.search(timestamp_pattern, raw_stat).group("timestamp")

            for port in port_list:
                # If port encountered first time
                if port not in stats_dict_list:
                    stats_dict_list[port] = [self.get_meta_item()]

                stats_dict_list[port].append({"timestamp": timestamp})

            for param in SsRunner.param_list:
                pattern = (
                    r"\s"
                    + re.escape(param)
                    + r"[\s:](?P<value>\w+\.?\w*(?:[\/\,]\w+\.?\w*)*)\s"
                )
                # result list stores all the string that is matched by the `pattern`
                param_value_list = [
                    value.group("value") for value in re.finditer(pattern, raw_stat)
                ]
                param_value = ""
                for i in range(len(param_value_list)):
                    param_value = param_value_list[i].strip()
                    # remove the (rate) units at the end and convert
                    if param_value.endswith("bps"):
                        param_value = self.convert_to(param_value)
                    try:
                        # RTT has both average and RTT deviation separated by a /
                        if param == "rtt":
                            avg_rtt = param_value.split("/")[0]
                            dev_rtt = param_value.split("/")[1]
                            stats_dict_list[port_list[i]][-1]["rtt"] = avg_rtt
                            stats_dict_list[port_list[i]][-1]["dev_rtt"] = dev_rtt
                        else:
                            stats_dict_list[port_list[i]][-1][param] = param_value
                    except TypeError:
                        pass

        destination_ip = self.destination_address.get_addr(with_subnet=False)
        SsResults.add_result(self.ns_id, {destination_ip: stats_dict_list})

    @staticmethod
    def convert_to(param_value, unit_out="Mbps"):
        """
        Converts parameter value to specified unit [For bit per second]
        """
        converter = {"bps": 1, "Kbps": 1e3, "Mbps": 1e6, "Gbps": 1e9}

        # Extract value and unit
        unit_in = re.sub(r"^\d*\.?\d*", "", param_value)
        extracted_param_value = float(re.sub(r"[A-Za-z]*", "", param_value))

        converted_param_value = str(
            extracted_param_value * converter[unit_in] / converter[unit_out]
        )
        return converted_param_value



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""Plot ss results"""

import logging
import matplotlib.pyplot as plt
from nest.experiment.interrupts import handle_keyboard_interrupt
from ..pack import Pack
from .common import simple_plot

logger = logging.getLogger(__name__)


def _extract_from_iperf3_flow(flow, node, dest_ip, local_port):
    """
    Extract information from flow data and convert it to
    conveniently plottable data format

    Parameters
    ----------
    flow : List
        List with timestamps and stats
    node : string
        Node from which ss results were obtained from
    dest_ip : string
        Destination ip address of the flow
    local_port : string
        Local port of the flow

    Returns
    -------
    (timestamp, sending_rate) or None
        Return timestamp and sending rate
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        logger.warning(
            "Flow from %s and port %s to destination ip %s doesn't have any parsed ss result.",
            node,
            dest_ip,
            local_port,
        )
        return None

    # First item is the "meta" item with user given information
    user_given_start_time = float(flow[0]["start_time"])
    destination_node = flow[0]["destination_node"]

    # "Bias" actual start_time in experiment with user given start time
    start_time = float(flow[1]["timestamp"]) - user_given_start_time

    timestamp = []
    sending_rate = []

    for data in flow[1:]:
        sending_rate.append(float(data["sending_rate"]))
        relative_time = float(data["timestamp"]) - start_time
        timestamp.append(relative_time)

    return {"destination_node": destination_node, "values": (timestamp, sending_rate)}


def _plot_iperf3_flow(flow, node, dest_ip, local_port):
    """
    Plot iperf3 stats of the flow

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    flow : List
        List with timestamps and stats
    node : str
        Node from which iperf3 results were obtained from
    dest_ip : str
        Destination ip address of the flow
    local_port : str
        Local port of the flow
    """
    data = _extract_from_iperf3_flow(flow, node, dest_ip, local_port)
    destination_node = data["destination_node"]
    values = data["values"]

    if values is None:
        return
    (timestamp, sending_rate) = values

    fig = simple_plot(
        "",
        timestamp,
        sending_rate,
        "Time (Seconds)",
        "Sending Rate (Mbps)",
        legend_string=f"{node} from port {local_port} to {destination_node} ({dest_ip})",
    )

    filename = f"sending_rate_{node}_to_{destination_node}({dest_ip}).png"
    Pack.dump_plot("iperf3", filename, fig)
    plt.close(fig)


# pylint: disable=too-many-locals
@handle_keyboard_interrupt
def plot_iperf3(parsed_data):
    """
    Plot statistics obtained from iperf3

    Parameters
    ----------
    exp_name : string
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from iperf3
    """

    # pylint: disable=too-many-nested-blocks
    for node in parsed_data:
        node_data = parsed_data[node]
        for connection in node_data:
            for dest_ip in connection:

                flow_data = connection[dest_ip]
                for local_port in flow_data:
                    flow = flow_data[local_port]
                    _plot_iperf3_flow(flow, node, dest_ip, local_port)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to Network"""

import logging
from nest.input_validator import input_validator
from nest.topology_map import TopologyMap
from .address import Address

logger = logging.getLogger(__name__)


class Network:
    """
    Abstraction for the network of interfaces.

    Attributes
    ----------
    address: str
        IP address of the network

    """

    current_network = None

    @input_validator
    def __init__(self, network_address: Address):
        """
        Constructor of Network.

        Parameters
        ----------
        network_address : str
            IP address of the network

        """

        self.net_address = network_address
        self.interface = []

        # Adding each network's object reference to the static list of networks in topology_map.
        TopologyMap.add_network(self)

    def __enter__(self):
        """
        Enter the context of this `Network`.

        """
        # Storing the currently executed network object to the static network variable.
        Network.current_network = self

    def __exit__(self, *args):
        """
        Exit the context of this `Network`.
        """
        Network.current_network = None

    # TODO: Handle the get interface request using the interface list

    def add_interface(self, _interface=None):
        """
        Adding interface to the network.

        Parameters
        ----------
        _interface : interface
            The interface which needs to be added to the Network.
        """
        self.interface.append(_interface)
        TopologyMap.decrement_orphan_interfaces()

    def __repr__(self):
        classname = self.__class__.__name__
        return f"{classname}({self.net_address!r})"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Necessary checks before running utilities on experiment"""

import functools
from time import sleep
import nest.config as config
import nest.global_variables as g_var
from nest.topology_map import TopologyMap
from nest.engine.ipv6_states import check_ipv6_states


def ipv6_dad_check(func):
    """
    If DAD is enabled, verifies the state of interfaces
    """

    @functools.wraps(func)
    def wrapper_dad_check(*args, **kwargs):
        """
        Wrapper function for DAD

        Parameters
        ----------
        *args : Non-Keyword Arguments
            passes variable number of arguments to a function
        **kwargs : Keyword Arguments
            passes keyworded, variable-length argument list to a function

        Returns
        -------
        func
            Function to be executed after wrapper is executed
        """

        if (
            g_var.IS_IPV6 is True
            and config.get_value("disable_dad") is not True
            and g_var.IS_DAD_CHECKED is not True
        ):

            namespaces = TopologyMap.get_namespaces()

            # Verifies if IPv6 states are addressable or not
            while True:
                status = check_ipv6_states(namespaces)

                # IPv6 state will be both in tentative and dadfailed together
                if status["dadfailed"][0] is True:
                    raise Exception(
                        "Duplicate address found "
                        f"at interface of node {status['dadfailed'][1]}."
                        "\nExiting ...."
                    )

                # Wait if IPv6 interface is in tentative state
                if status["tentative"] is True:
                    sleep(1)
                else:
                    break
            g_var.IS_DAD_CHECKED = True
        return func(*args, **kwargs)

    return wrapper_dad_check



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to switch creation in topology"""

import logging

from nest import engine
from nest.topology.node import Node

logger = logging.getLogger(__name__)


class Switch(Node):
    """
    Abstraction for a switch.

    Attributes
    ----------
    name: str
        User given name for the switch.
    """

    def __init__(self, name):
        """
        Create a switch with given `name` inside a 'Node'.

        A namespace is created by inheriting the "Node" class and unique id is given
        to the namespace. Then a switch with the same name and unique id as namespace
        is created inside the created namespace.

        Parameters
        ----------
        name: str
            The name of the switch to be created
        """
        super().__init__(name)
        engine.create_switch(self.id, self.id)
        engine.set_switch_mode(self.id, "up")

    def _add_interface(self, interface):
        """
        Add `interface` to `Switch`

        Parameters
        ----------
        interface: Interface
            `Interface` to be added to `Switch`
        """
        super()._add_interface(interface)
        engine.add_int_to_switch(self.id, interface.id)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Runs and provides RTT output from ping command"""

import re
from time import sleep
from functools import partial
from nest.experiment.interrupts import handle_keyboard_interrupt
from .runnerbase import Runner
from ..results import PingResults
from ...engine.ping import run_exp_ping


class PingRunner(Runner):
    """
    Runs ping command and parses statistics from its output

    Attributes
    ----------
    ns_id : str
        network namespace to run netperf from
    destination_ip : str
        IP address of the destination namespace
    start_time : num
        time at which netperf is to run
    run_time : num
        total time to run netperf for
    """

    # pylint: disable=too-many-arguments
    def __init__(self, ns_id, destination_ip, start_time, run_time, dst_ns):
        """
        Constructor to initialize ping runner

        Parameters
        ----------
        ns_id : str
            network namespace to run ping from
        destination_ip : str
            IP address of the destination namespace
        start_time : num
            time at which netperf is to run
        run_time : num
            total time to run netperf for
        dst_ns : str
            destination network namespace of ping
        """
        super().__init__(ns_id, start_time, run_time, destination_ip, dst_ns)

    # pylint: disable=arguments-differ
    def run(self):
        """
        Runs ping at t=`self.start_time`
        """
        if self.start_time > 0:
            sleep(self.start_time)

        super().run(
            partial(
                run_exp_ping,
                self.ns_id,
                self.destination_address.get_addr(with_subnet=False),
                self.run_time,
                self.destination_address.is_ipv6(),
            ),
            error_string_prefix="Collecting latency",
        )

    @handle_keyboard_interrupt
    def parse(self):
        """
        parses the RTT from `self.out`
        """
        self.out.seek(0)  # rewind to start of the temp file
        raw_stats = self.out.read().decode()

        pattern = r"\[(?P<timestamp>\d+\.\d+)\].*time=(?P<rtt>\d+(\.\d+)?)"
        timestamps_and_rtts = [
            (match.group("timestamp"), match.group("rtt"))
            for match in re.finditer(pattern, raw_stats)
        ]

        # List storing collected stats
        # First item as "meta" item with user given information
        stats_list = [self.get_meta_item()]

        for (timestamp, rtt) in timestamps_and_rtts:
            stats_list.append({"timestamp": timestamp, "rtt": rtt})

        stats_dict = {self.destination_address.get_addr(with_subnet=False): stats_list}

        PingResults.add_result(self.ns_id, stats_dict)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
Helper methods for handling types from typing module.

NOTE: This module uses private members of typing module
for backward compatibility with python 3.7 and older versions.
"""

import typing


def is_optional(field):
    """
    Returns if the `field` type is of Optional type

    Parameters
    ----------
    field: typing
        The type to be checked

    Returns
    -------
    bool
        Returns True if `field` is optional
    """
    if hasattr(field, "__args__") and hasattr(field, "__origin__"):
        if (
            # pylint: disable=comparison-with-callable
            field.__origin__ == typing.Union
            and len(field.__args__) == 2
            and field.__args__[-1] is type(None)
        ):
            return True

    return False


def get_inner_type_of_optional_field(field):
    """
    Get inner type inside optional field
    """
    return field.__args__[0]


def is_list(field):
    """
    Checks if `field` type is a list.
    """
    if hasattr(field, "__args__") and hasattr(field, "__origin__"):
        if field.__origin__ == list:
            return True
    return False



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
Helper methods to handle interrupts from the user.

For eg., KeyboardInterrupt
"""

import functools
import logging

logger = logging.getLogger(__name__)


def handle_keyboard_interrupt(func):
    """
    Decorator for handling keyboard interrupt.
    It is assumed that this decorator is used in subprocesses.
    (the debug message is formulated accordingly)
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            result = func(*args, **kwargs)
        except KeyboardInterrupt:
            logger.debug(
                "Running %s: Received KeyboardInterrupt, "
                "hence shutting down the process gracefully.",
                func.__qualname__,
            )
        return result

    return wrapper



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
NeST keeps track of all global variables here

WARNING: Avoid adding new variables in this file
         until its the only last resort
"""

# Keeps track if DAD checked or not,
# whenever new topology is defined
IS_DAD_CHECKED = False

# Keeps track if address is IPv6 or not for DAD check
IS_IPV6 = False



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handles OSPF related functionalities"""

import random
from functools import partial
from nest.engine.dynamic_routing import run_ospfd
from nest.routing.route_daemons import RoutingDaemonBase
import nest.config as config


class Ospf(RoutingDaemonBase):
    """
    Handles OSPF related functionalities.
    """

    def __init__(self, router_ns_id, interfaces, conf_dir, **kwargs):
        super().__init__(router_ns_id, interfaces, "ospfd", conf_dir, **kwargs)

    def create_basic_config(self):
        """
        Creates a file with basic configuration for OSPF.
        Use base `add_to_config` directly for more complex configurations
        """
        if self.ipv6:
            for interface in self.interfaces:
                self.add_to_config(f"interface {interface.id}")
                # send hello packets every 1 second for faster convergence
                self.add_to_config("ipv6 ospf6 hello-interval 1")

            self.add_to_config("router ospf6")

            # Generates random router-id in A.B.C.D format
            router_id = ".".join(
                map(str, (random.randint(0, 255) for _ in range(0, 4)))
            )
            # for quagga
            if config.get_value("routing_suite") == "quagga":
                self.add_to_config(f"router-id {router_id}")
            # for frr
            else:
                self.add_to_config(f"ospf6 router-id {router_id}")
            for interface in self.interfaces:
                self.add_to_config(f" interface {interface.id} area 0.0.0.0")
        else:
            for interface in self.interfaces:
                self.add_to_config(f"interface {interface.id}")
                # send hello packets every 1 second for faster convergence
                self.add_to_config("ip ospf hello-interval 1")

            self.add_to_config("router ospf")
            self.add_to_config(
                f"ospf router-id {self.interfaces[0].address.get_addr(with_subnet=False)}"
            )
            for interface in self.interfaces:
                self.add_to_config(
                    f" network {interface.address.get_subnet()} area 0.0.0.0"
                )

        if self.log_file is not None:
            self.add_to_config(f"log file {self.log_file}")

        self.create_config()

    def run(self):
        """
        Runs the ospfd command
        """
        super().run(
            engine_func=partial(
                run_ospfd, self.router_ns_id, self.conf_file, self.pid_file, self.ipv6
            )
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Package all results into a folder"""

import os
import time
import shutil

from nest.user import User


class Pack:
    """Handles packaging results"""

    FOLDER = ""

    @staticmethod
    def init(exp_name):
        """
        Create a folder with format 'testname(time)_dump'

        Parameters
        ----------
        exp_name : str
            Name of experiment
        """
        timestamp = time.strftime("%d-%m-%Y-%H:%M:%S")
        Pack.FOLDER = "{exp_name}({timestamp})_dump".format(
            exp_name=exp_name, timestamp=timestamp
        )
        os.mkdir(Pack.FOLDER)
        Pack.set_owner(Pack.FOLDER)

    @staticmethod
    def dump_file(filename, content):
        """
        Dump a file into Pack.FOLDER

        Parameters
        ----------
        filename : str
            Name of file
        content : str
            Content to be stored in file
        """
        path = os.path.join(Pack.FOLDER, filename)
        with open(path, "w") as file:
            file.write(content)
        Pack.set_owner(path)

    @staticmethod
    def dump_plot(subfolder, filename, fig):
        """
        Dump a plot into Pack.FOLDER

        Parameters
        ----------
        subfolder : str
            Subfolder to which plot belongs to
        filename : str
            Name of plot
        fig : matplotlib.pyplot.fig
            Plot figure to be stored in file
        """
        Pack.create_subfolder(subfolder)
        path = os.path.join(Pack.FOLDER, subfolder, filename)
        fig.savefig(path)
        Pack.set_owner(path)

    @staticmethod
    def create_subfolder(subfolder):
        """
        Create subfolder if it already doesn't exist

        Parameters
        ----------
        subfolder : str
            Subfolder name
        """
        path = os.path.join(Pack.FOLDER, subfolder)
        if not os.path.isdir(path):
            os.mkdir(path)
            Pack.set_owner(path)

    @staticmethod
    def set_owner(path):
        """
        Change file permission to that of the
        actual user running (For example, user running
        sudo)

        Parameters
        ----------
        path : str
            Path of file
        """

        user_id = User.user_id
        group_id = User.group_id
        os.chown(path, user_id, group_id)

    @staticmethod
    def compress():
        """Compress Pack.FOLDER into a tar archive"""
        # TODO

    @staticmethod
    def copy_files(src_path, dst_path=None):
        """
        Copies file from source to stats folder

        Parameters
        ----------
        src_path : str
            Path of source file
        dst_path : str
            Relative Path of destination inside the stats folder
        """

        if dst_path is None:
            # Default destination path is stats folder
            dst_path = Pack.FOLDER
        else:
            dst_path = os.path.join(Pack.FOLDER, dst_path)

        shutil.copy(src_path, dst_path)
        filename = os.path.join(dst_path, os.path.basename(src_path))
        Pack.set_owner(filename)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from routing sub-package"""

import unittest
from nest import config
from nest.topology_map import TopologyMap
from nest.topology import Node, connect
from nest.routing.routing_helper import RoutingHelper
from nest.clean_up import delete_namespaces

# pylint: disable=missing-docstring


class TestQuagga(unittest.TestCase):

    # pylint: disable=invalid-name
    def setUp(self):
        self.n0 = Node("n0")
        self.n1 = Node("n1")
        self.r0 = Node("r0")
        self.r1 = Node("r1")
        self.r0.enable_ip_forwarding()
        self.r1.enable_ip_forwarding()

        ### Create interfaces and connect nodes and routers ###

        (eth_p1r1, eth_r1p1) = connect(self.n0, self.r0, "eth-n1r1-0", "eth-r1n1-0")
        (eth_r1r2, eth_r2r1) = connect(self.r0, self.r1, "eth-r1r2-0", "eth-r2r1-0")
        (eth_r2p2, eth_p2r2) = connect(self.r1, self.n1, "eth-r2n2-0", "eth-n2r2-0")

        ### Assign addresses to interfaces ###

        eth_p1r1.set_address("10::1:1/122")
        eth_r1p1.set_address("10::1:2/122")

        eth_r1r2.set_address("10::2:2/122")
        eth_r2r1.set_address("10::2:3/122")

        eth_r2p2.set_address("10::3:3/122")
        eth_p2r2.set_address("10::3:4/122")

        config.set_value("routing_suite", "quagga")  # Use quagga

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def test_routing_helper(self):

        RoutingHelper("rip").populate_routing_tables()

        status = self.n0.ping("10::3:4", verbose=False)
        self.assertTrue(status)

        status = self.n1.ping("10::1:1", verbose=False)
        self.assertTrue(status)

    def test_ospf(self):
        RoutingHelper("ospf").populate_routing_tables()

        status = self.n0.ping("10::3:4", verbose=False)
        self.assertTrue(status)

        status = self.n1.ping("10::1:1", verbose=False)
        self.assertTrue(status)

    def test_isis(self):
        RoutingHelper("isis").populate_routing_tables()

        status = self.n0.ping("10::3:4", verbose=False)
        self.assertTrue(status)

        status = self.n1.ping("10::1:1", verbose=False)
        self.assertTrue(status)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

""" exectute iterator scripts """
from .exec import exec_exp_commands

# pylint: disable=too-many-arguments
def run_ss(
    ns_id, iterator, destination_ip, duration, ss_filter, start_time, ipv6, out, err
):
    """
    Executes the ss iterator script

    Parameters
    ----------
    ns_id : str
        namespace name
    iterator : str
        absolute path of the ss iterator script
    destination_ip : str
        IP address of the destination namespace
    duration : num
        total time to run ss for
    start_time : num
        time at which ss is to be run
    ss_filter : str
        filter to remove unnecessary output from ss
    ipv6 : bool
        determines if destination_ip is ipv4/ipv6
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    if ipv6:
        return exec_exp_commands(
            f"ip netns exec {ns_id} /bin/bash {iterator} [{destination_ip}] \
                                    {duration} {ss_filter}  {start_time}",
            stdout=out,
            stderr=err,
        )

    return exec_exp_commands(
        f"ip netns exec {ns_id} /bin/bash {iterator} {destination_ip} \
                                {duration} {ss_filter}  {start_time}",
        stdout=out,
        stderr=err,
    )


# pylint: disable=too-many-arguments
def run_tc(ns_id, iterator, dev, duration, out, err):
    """
    Executes the tc iterator script

    Parameters
    ----------
    ns_id : str
        network namespace to run tc from
    iterator : str
        absolute path of the tc iterator script
    dev : str
        dev id to collect tc stats from
    duration : num
        total time to run tc for
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    return exec_exp_commands(
        f"ip netns exec {ns_id} /bin/bash {iterator} {dev} {duration}",
        stdout=out,
        stderr=err,
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Base class for Routing daemons.
"""

from abc import ABC, abstractmethod
import io
import logging
import shutil
from nest.engine.dynamic_routing import supports_dynamic_routing
from nest import config
from nest.logging_helper import DepedencyCheckFilter, DuplicateRoutingLogsFilter

# pylint: disable=too-many-instance-attributes


class RoutingDaemonBase(ABC):
    """
    Abstract class for Dynamic routing related processes.
    This class should be inherited for adding other daemons

    Use `add_to_config` to sequentially add daemon related commands
    for the config file.

    Finally call `run` to create the config file and  run the daemon.

    Note:
    (i): If you're using `RoutingHelper`, config files are created at the /tmp
    directory.
    (ii): Daemons may not work as expected if a config file is not created

    Attributes
    ----------
    conf: file-like
        In memory stream to temporarily hold configuration
    router_ns_id : str
        Router namespace id
    daemon : str
        daemon to run(one of ['zebra', 'ospf', 'isis'])
    conf_file : str
        config file path
    pid_file : str
        pid file path for the daemon process
    interfaces : List[Interface]
        interfaces present in the router
    """

    def __init__(self, router_ns_id, interfaces, daemon, conf_dir, **kwargs):
        """
        Parameters
        ----------
        conf_dir : str
            Directory to store config files
        **kwargs
            Key worded arguments for other daemon specific parameters
            ``log_dir``:
                Directory to store log files. (`str`)
        """
        self.logger = logging.getLogger(__name__)
        self.daemon = daemon
        if not any(
            isinstance(filter, DepedencyCheckFilter) for filter in self.logger.filters
        ):
            # Duplicate filter is added to avoid logging of same error
            # message incase any of the routing daemon is not installed
            self.logger.addFilter(DepedencyCheckFilter())

        if not any(
            isinstance(filter, DuplicateRoutingLogsFilter)
            for filter in self.logger.filters
        ):
            self.logger.addFilter(DuplicateRoutingLogsFilter())

        if not supports_dynamic_routing(daemon):
            self.handle_dependecy_error()

        self.conf = io.StringIO()
        self.router_ns_id = router_ns_id
        self.conf_file = f"{conf_dir}/{self.router_ns_id}_{daemon}.conf"
        self.pid_file = f"{conf_dir}/{self.router_ns_id}_{daemon}.pid"
        self.log_file = None
        if kwargs["log_dir"] is not None:
            self.logger.info(
                "%s logging enabled. Log files can found in %s directory",
                config.get_value("routing_suite"),
                kwargs["log_dir"],
            )
            self.log_file = f"{kwargs['log_dir']}/{self.router_ns_id}_{daemon}.log"
        self.interfaces = interfaces
        self.ipv6 = interfaces[0].address.is_ipv6()

    @abstractmethod
    def create_basic_config(self):
        """
        Created minimum configuration for `daemon`
        """

    @abstractmethod
    def run(self, engine_func):
        """
        Run the `daemon` along with its config file
        """
        self.create_config()
        engine_func()

    def add_to_config(self, command):
        """
        Add a line to `self.conf`

        Parameters
        ---------
        command : str
            command to add to config file
        """
        self.conf.write(f"{command}\n")

    def create_config(self):
        """
        Creates config file on disk from `self.conf`
        """
        with open(self.conf_file, "w") as conf:
            shutil.chown(self.conf_file, user=config.get_value("routing_suite"))
            self.conf.seek(0)
            shutil.copyfileobj(self.conf, conf)

    def handle_dependecy_error(self):
        """
        Default error when routing daemon is not present
        """
        self.logger.error("%s not found. Routes may not be added properly", self.daemon)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
This module contains methods to kill any running processes in namespaces
and delete all namespaces after the experiment is complete.
"""

import atexit
import logging
from nest.network_utilities import ipv6_dad_check
from . import config

from . import engine
from .topology_map import TopologyMap

logger = logging.getLogger(__name__)


def kill_processes():
    """
    Kill any running processes in namespaces
    """
    for namespace in TopologyMap.get_namespaces():
        engine.kill_all_processes(namespace["id"])


@atexit.register
@ipv6_dad_check
def delete_namespaces():
    """
    Delete all the newly generated namespaces
    """
    namespaces = TopologyMap.get_namespaces()

    if config.get_value("delete_namespaces_on_termination"):
        for namepspace in namespaces:
            engine.delete_ns(namepspace["id"])
        logger.info("Cleaned up environment!")
    else:
        logger.info("Namespaces not deleted")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from mpls static routing"""

import subprocess
import unittest
from nest.topology import Node, connect
from nest.clean_up import delete_namespaces

# pylint: disable=missing-docstring


class TestStaticMPLS(unittest.TestCase):
    # pylint: disable=invalid-name
    def setUp(self):

        self.assertTrue(
            "mpls_iptunnel" in subprocess.check_output("lsmod").decode(),
            "Couldn't load mpls",
        )

        self.n0 = Node("n0")
        self.n1 = Node("n1")
        self.r0 = Node("r0")
        self.r1 = Node("r1")
        self.r0.enable_ip_forwarding()
        self.r1.enable_ip_forwarding()

    def tearDown(self):
        delete_namespaces()

        ### Create interfaces and connect nodes and routers ###

    def test_routing(self):
        (eth_n0_r0, eth_r0_n0) = connect(self.n0, self.r0, "eth-n0_r0", "eth-r0_n0")
        (eth_r0_r1, eth_r1_r0) = connect(self.r0, self.r1, "eth-r0r1-0", "eth-r1r0-0")
        (eth_r1_n1, eth_n1_r1) = connect(self.r1, self.n1, "eth-r1n1-0", "eth-n1r1-0")

        ### Assign addresses to interfaces ###
        eth_n0_r0.set_address("10.0.1.1/24")
        eth_r0_n0.set_address("10.0.1.2/24")

        eth_r0_r1.set_address("10.0.2.2/24")
        eth_r1_r0.set_address("10.0.2.3/24")

        eth_r1_n1.set_address("10.0.3.3/24")
        eth_n1_r1.set_address("10.0.3.4/24")

        # Enable MPLS on all interfaces
        eth_n0_r0.enable_mpls()
        eth_r0_n0.enable_mpls()

        eth_r0_r1.enable_mpls()
        eth_r1_r0.enable_mpls()

        eth_r1_n1.enable_mpls()
        eth_n1_r1.enable_mpls()

        # Add routes(mpls family)
        self.n0.add_route_mpls_push("10.0.3.0/24", eth_r0_n0.get_address(), 101)
        self.r0.add_route_mpls_switch(101, eth_r1_r0.get_address(), 102)
        self.r1.add_route_mpls_pop(102, eth_n1_r1.get_address())

        self.n1.add_route_mpls_push("10.0.1.0/24", eth_r1_n1.get_address(), 201)
        self.r1.add_route_mpls_switch(201, eth_r0_r1.get_address(), 202)
        self.r0.add_route_mpls_pop(202, eth_n0_r0.get_address())

        status = self.n0.ping("10.0.3.4", verbose=True)
        self.assertTrue(status)

        status = self.n1.ping("10.0.1.1", verbose=True)
        self.assertTrue(status)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to Address Helper"""

import logging
from nest.topology_map import TopologyMap
from .address import Subnet

# pylint: disable=too-few-public-methods

logger = logging.getLogger(__name__)


class AddressHelper:
    """
    Abstraction for automatic address assignment to interfaces of the network.
    """

    @staticmethod
    def assign_addresses(network=None):
        """
        Assignment of addresses to the interfaces.

        Parameters
        ----------
        network : Network
            Assigning the addresses to each interfaces of all network.
            It will assign addresses to specific network if network object is mentioned.
        """
        if network is None:
            for net in TopologyMap.list_of_network:
                AddressHelper.__assign_addresses_to_network(net)

            # Warning for the assignment of the addresses to the interfaces,
            # which do not belong to any network
            if TopologyMap.orphan_interfaces != 0:
                logger.warning(
                    "Interfaces not part of some network. The addresses to these orphaned "
                    "interfaces should be manually added.",
                )
        else:
            AddressHelper.__assign_addresses_to_network(network)

    @staticmethod
    def __assign_addresses_to_network(network):
        """
        Assignment of addresses to all interfaces of specific network.

        Parameters
        ----------
        network : Network
            Assigning the addresses to each interfaces of the given network object.
        """
        _interface = []
        _interface = network.interface

        _net_address = Subnet(network.net_address)

        for inter in _interface:
            inter.set_address(_net_address.get_next_addr())



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""IP address commands"""

from .exec import exec_subprocess


def assign_ip(host_name, dev_name, ip_address):
    """
    Assigns ip address to interface

    Parameters
    ----------
    host_name : str
        name of the host namespace
    dev_name : str
        name of the interface
    ip_address : str
        ip address to be assigned to the interface
    """
    # TODO: Support for IPv6
    exec_subprocess(
        f"ip netns exec {host_name} ip address add {ip_address} dev {dev_name}"
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Runs tc command and parses qdisc stats
from its output
"""

import re
import json
import os
from functools import partial
from time import strptime, strftime
from nest.experiment.interrupts import handle_keyboard_interrupt
from .runnerbase import Runner
from ..results import TcResults
from ...topology_map import TopologyMap
from ...engine.tc import get_tc_version
from ...engine.iterators import run_tc


class TcRunner(Runner):
    """
    Runs tc command and stores and parses the output

    Attributes
    ----------
    iterator : str
        absolute path of the tc iterator script
    old_kernel_version : str
        minimum kernel version required
    new_kernel_version : str
        kernel version from which tc has concrete JSON support
    ns_id : str
        network namespace to run tc from
    dev : str
        dev id to collect tc stats from
    run_time : num
        total time to run tc for
    """

    iterator = os.path.realpath(os.path.dirname(__file__)) + "/iterators/tc.sh"

    # tc versions are in date formatted
    # TODO: move these to a config file

    # NOTE: Both the below versions are in OLD tc version format
    # Refer docstring of `check_tc_version_format`

    # Minimum version of tc required by NeST
    MINIMUM_SUPPORTED_VERSION = strptime("20180129", "%Y%m%d")

    # tc version with good JSON support for displaying stats
    JSON_SUPPORTED_VERSION = strptime("20190319", "%Y%m%d")

    # Qdiscs supported prior to good JSON support in tc
    PRIOR_JSON_QDISCS_SUPPORTED = ["codel", "fq_codel", "pie"]

    def __init__(self, ns_id, dev, qdisc, run_time):
        """
        Constructor to initialize tc runner

        Parameters
        ----------
        ns_id : str
            network namespace to run tc from
        dev : str
            dev id to collect tc stats from
        qdisc : str
            qdisc name [eg. 'codel', 'pie']
        run_time : num
            total time to run tc for
        """
        self.dev = dev
        self.qdisc = qdisc

        # Start parsing from 0s
        super().__init__(ns_id, 0, run_time)

        # Tc version check
        self.version_check()

    def version_check(self):
        """
        Check the tc version and throw exception if any
        unsupported task is requested from NeST
        """
        tc_version_format = self.check_tc_version_format()

        if tc_version_format == "old_version_format":
            cur_tc_version = self.parsed_tc_version()

            if cur_tc_version < TcRunner.MINIMUM_SUPPORTED_VERSION:
                # TODO: Not sure if it's the right exception to raise
                version = "iproute2-ss" + strftime(
                    "%y%m%d", TcRunner.MINIMUM_SUPPORTED_VERSION
                )
                raise Exception(
                    f"NeST does not support qdisc parsing for tc version below "
                    f"{version}"
                )

            if cur_tc_version < TcRunner.JSON_SUPPORTED_VERSION:
                if self.qdisc not in TcRunner.PRIOR_JSON_QDISCS_SUPPORTED:

                    version = "iproute2-ss" + strftime(
                        "%y%m%d", TcRunner.JSON_SUPPORTED_VERSION
                    )
                    raise Exception(
                        f"NeST does not support {self.qdisc} qdisc parsing for tc "
                        f"version below {version}. Supported "
                        f"qdiscs are {TcRunner.PRIOR_JSON_QDISCS_SUPPORTED}"
                    )

    def run(self):
        """
        Runs the tc iterator
        """
        super().run(
            partial(run_tc, self.ns_id, TcRunner.iterator, self.dev, self.run_time),
            error_string_prefix="Collecting qdisc stats",
        )

    def get_qdisc_specific_params(self):
        """
        Parameters to be obtained for a specific qdisc

        Returns
        -------
        dict:
            qdisc keyed list of parameters to parsed
        """
        qdisc_param = {
            "codel": ["count", "lastcount", "ldelay", "drop_next"],
            "fq_codel": ["maxpacket", "drop_overlimit", "new_flow_count"],
            "pie": ["prob", "delay", "avg_dq_rate"],
        }
        return qdisc_param

    def get_qdisc_re(self):
        """
        Compile regular expression for parsing qdisc specific parameters

        Returns
        -------
        dict
            qdisc keyed regular expression
        """
        qdisc_re = {}
        qdisc_re["codel"] = re.compile(
            r"count (?P<count>\d+) "
            r"lastcount (?P<lastcount>\d+) "
            r"ldelay (?P<ldelay>[0-9\.]+[mu]?s) "
            r"(?P<dropping>dropping)? ?"
            r"drop_next (?P<drop_next>-?[0-9\.]+[mu]?s)"
        )
        qdisc_re["fq_codel"] = re.compile(
            r"maxpacket (?P<maxpacket>\d+) "
            r"drop_overlimit (?P<drop_overlimit>\d+) "
            r"new_flow_count (?P<new_flow_count>\d+) "
        )
        qdisc_re["pie"] = re.compile(
            r"prob (?P<prob>[0-9\.]+) "
            r"delay (?P<delay>[0-9\.]+[mu]?s) "
            r"avg_dq_rate (?P<avg_dq_rate>\d+)"
        )
        return qdisc_re

    def repl(self, match):
        """
        Called by re.sub() for every match

        Parameters
        ----------
        match : Match
            Match object captured by the pattern

        Returns
        -------
        str
            string to replace the matched string
        """
        if match.group(1):
            if match.group(1).endswith(","):
                value = repr(match.group(1))
                return ':"{}",'.format(value)

            value = repr(match.group(1))
            return ':"{}"'.format(value)
        return ""

    def clean_json(self, stats):
        """
        JSON formatted tc stats with invalid JSON keys
        and values are removed or fixed

        Parameters
        ----------
        stats : str
            unclean json formatted stats

        Returns
        -------
        str
            fixed json stats
        """
        # pattern to remove the options key
        options_pattern = r'"options":(\s)*{(.|\n)*?},'
        stats = re.sub(options_pattern, "", stats)

        # pattern to enclose all the values with " "
        value_pattern = r":(\s(\w|\s|\n|\.)+,?)"
        stats = re.sub(value_pattern, self.repl, stats)
        return stats

    def parsing_helper_before_good_json_support(self, raw_stats, qdisc_param, qdisc_re):
        """
        Parsing tc command on Linux kernel versions
        4.15.0 to 5.4

        Parameters
        ----------
        raw_stats : list(str)
            list of stats collected at each iteration
        qdisc_param : dict
            parameters to be obtained for a specific qdisc
        qdisc_re : dict
            regular expression for parsing qdisc specific parameters

        Returns
        -------
        dict
            handle keyed dict with list of stats
        """

        aggregate_stats = {}

        for raw_stat in raw_stats[:-1]:
            timestamp_pattern = r"timestamp:(?P<timestamp>\d+\.\d+)"
            timestamp = re.search(timestamp_pattern, raw_stat).group("timestamp")
            raw_stat = re.sub(timestamp_pattern, "", raw_stat)
            raw_stat = json.loads(self.clean_json(raw_stat))
            stats_dict = {}
            for qdisc_stat in raw_stat:
                qdisc = qdisc_stat["kind"]
                if (
                    qdisc in TcRunner.PRIOR_JSON_QDISCS_SUPPORTED
                    and qdisc == self.qdisc
                ):
                    handle = qdisc_stat["handle"]
                    if handle not in aggregate_stats:
                        aggregate_stats[handle] = []
                    qdisc_stat = qdisc_stat["qlen"]
                    search_obj = qdisc_re[qdisc].search(qdisc_stat)
                    stats_dict["timestamp"] = str(timestamp)
                    stats_dict["kind"] = qdisc
                    for param in qdisc_param[qdisc]:
                        stats_dict[param] = search_obj.group(param)
                    aggregate_stats[handle].append(stats_dict)
        return aggregate_stats

    def parsing_helper(self, raw_stats):
        """
        Parsing tc command on Linux kernel version
        5.5 and above

        Parameters
        ----------
        raw_stats : list(str)
            list of stats collected at each iteration

        Returns
        -------
        dict
            handle keyed dict with list of stats
        """
        aggregate_stats = {}
        for raw_stat in raw_stats[:-1]:
            timestamp_pattern = r"timestamp:(?P<timestamp>\d+\.\d+)"
            timestamp = re.search(timestamp_pattern, raw_stat).group("timestamp")
            raw_stat = re.sub(timestamp_pattern, "", raw_stat)
            raw_stat = json.loads(raw_stat)
            stats_dict = {}
            for qdisc_stat in raw_stat:
                qdisc = qdisc_stat["kind"]
                if qdisc == self.qdisc:  # To ignore the HTB qdisc
                    handle = qdisc_stat["handle"]
                    if handle not in aggregate_stats:
                        aggregate_stats[handle] = []

                    stats_dict["timestamp"] = str(timestamp)
                    stats_dict.update(qdisc_stat)
                    stats_dict.pop("handle", None)
                    stats_dict.pop("options", None)
                    stats_dict.pop("parent", None)

                    aggregate_stats[handle].append(stats_dict)
        return aggregate_stats

    def check_tc_version_format(self):
        """
        iproute2 changed its versioning format after v5.8.0

        For example, below are the output of `tc -V` for versions near
        v5.8.0 (this list was compiled by checking the iproute2 git repository):

        main   - tc utility, iproute2-5.8.0
        v5.8.0 - tc utility, iproute2-v5.7.0-77-gb687d1067169
        v5.7.0 - tc utility, iproute2-ss200602
        v5.6.0 - tc utility, iproute2-ss200330

        So the function will return 'new_version_format' for versions
        after v5.7.0, else the function will return 'old_version_format'
        """
        old_version_format = "tc utility, iproute2-ss[0-9]{6}\n"
        tc_version = get_tc_version()

        if re.search(old_version_format, tc_version):
            return "old_version_format"

        return "new_version_format"

    def parsed_tc_version(self):
        """
        [For OLD tc format, refer docstring of `check_tc_version_format`]

        Parses the current tc version

        Returns
        -------
        struct_time
            current tc version as date
        """
        cur_tc_version = get_tc_version()
        cur_tc_version = "20" + cur_tc_version.split(" ")[-1][-7:].strip()
        return strptime(cur_tc_version, "%Y%m%d")

    @handle_keyboard_interrupt
    def parse(self):
        """
        Parses the required data from tc-qdisc output
        """

        self.out.seek(0)  # rewind to start of the temp file

        # See `iterators/tc.sh` for output format
        raw_stats = self.out.read().decode().split("---")
        aggregate_stats = {}

        tc_version_format = self.check_tc_version_format()

        if tc_version_format == "new_version_format":
            aggregate_stats = self.parsing_helper(raw_stats)

        elif tc_version_format == "old_version_format":
            cur_tc_version = self.parsed_tc_version()

            # tc produces different JSON output format
            # based on the version
            if cur_tc_version >= TcRunner.JSON_SUPPORTED_VERSION:
                aggregate_stats = self.parsing_helper(raw_stats)

            elif cur_tc_version >= TcRunner.MINIMUM_SUPPORTED_VERSION:
                qdisc_param = self.get_qdisc_specific_params()
                qdisc_re = self.get_qdisc_re()
                aggregate_stats = self.parsing_helper_before_good_json_support(
                    raw_stats, qdisc_param, qdisc_re
                )

        # Store parsed results
        dev_name = TopologyMap.get_interface(self.ns_id, self.dev)["name"]
        TcResults.add_result(self.ns_id, {dev_name: aggregate_stats})



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Routing commands"""
from os import path
from nest.engine.util import is_dependency_installed
from .. import config
from .exec import exec_subprocess


# Path to routing_suite daemon binaries
FRR_DAEMONPATH = "/usr/lib/frr/"


def run_zebra(ns_id, conf_file, pid_file):
    """
    Runs the zebra daemon

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    if config.get_value("routing_suite") == "frr":
        cmd = f"ip netns exec {ns_id} {FRR_DAEMONPATH}zebra --config_file {conf_file} \
                --pid_file {pid_file} --retain --daemon -N {ns_id}"
    else:
        cmd = f"ip netns exec {ns_id} zebra --config_file {conf_file} \
                --pid_file {pid_file} --retain --daemon"
    exec_subprocess(cmd)


def run_ripd(ns_id, conf_file, pid_file, ipv6):
    """
    Runs the ripd daemon

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    if config.get_value("routing_suite") == "frr":
        if ipv6:
            cmd = f"ip netns exec {ns_id} {FRR_DAEMONPATH}ripngd --config_file {conf_file} \
                    --pid_file {pid_file} --daemon -N {ns_id}"
        else:
            cmd = f"ip netns exec {ns_id} {FRR_DAEMONPATH}ripd --config_file {conf_file} \
                    --pid_file {pid_file} --daemon -N {ns_id}"
    else:
        if ipv6:
            cmd = f"ip netns exec {ns_id} ripngd --config_file {conf_file} \
                    --pid_file {pid_file} --retain --daemon"
        else:
            cmd = f"ip netns exec {ns_id} ripd --config_file {conf_file} \
                    --pid_file {pid_file} --retain --daemon"
    exec_subprocess(cmd)


def run_ospfd(ns_id, conf_file, pid_file, ipv6):
    """
    Runs the ospfd daemon

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    if config.get_value("routing_suite") == "frr":
        if ipv6:
            cmd = f"ip netns exec {ns_id} {FRR_DAEMONPATH}ospf6d --config_file {conf_file} \
                --pid_file {pid_file} --daemon -N {ns_id}"
        else:
            cmd = f"ip netns exec {ns_id} {FRR_DAEMONPATH}ospfd --config_file {conf_file} \
                --pid_file {pid_file} --daemon -N {ns_id}"
    else:
        if ipv6:
            cmd = f"ip netns exec {ns_id} ospf6d --config_file {conf_file} \
            --pid_file {pid_file} --daemon"
        else:
            cmd = f"ip netns exec {ns_id} ospfd --config_file {conf_file} \
                --pid_file {pid_file} --daemon"
    exec_subprocess(cmd)


def run_isisd(ns_id, conf_file, pid_file):
    """
    Runs the isisd daemon

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    if config.get_value("routing_suite") == "frr":
        cmd = f"ip netns exec {ns_id} {FRR_DAEMONPATH}isisd --config_file {conf_file} \
            --pid_file {pid_file} --daemon -N {ns_id}"
    else:
        cmd = f"ip netns exec {ns_id} isisd --config_file {conf_file} \
            --pid_file {pid_file} --daemon"
    exec_subprocess(cmd)


def run_ldpd(ns_id, conf_file, pid_file):
    """
    Runs the ldp daemon (MPLS)
    Requires frr routing_suite

    Parameters
    ----------
    ns_id : str
        namespace of the router
    conf_file : str
        path to config file
    pid_file : str
        path to pid file
    """
    if config.get_value("routing_suite") == "frr":
        cmd = f"ip netns exec {ns_id} /usr/lib/frr/ldpd --config_file {conf_file} \
            --pid_file {pid_file} --daemon -N {ns_id}"
        exec_subprocess(cmd)
    else:
        raise Exception("Ldp requires Frrouting")


def supports_dynamic_routing(daemon):
    """
    Checks whether frr/quagga is installed

    Parameters
    ----------
    daemon : str
        routing daemon

    Returns
    -------
    bool
        true if frr/quagga is installed
    """

    if config.get_value("routing_suite") == "quagga":
        return is_dependency_installed(daemon)

    if config.get_value("routing_suite") == "frr":
        # /usr/lib/ is the default installation path for frr which is not in PATH.
        # This results in `is_dependency_installed` always returning false for frr, hence we
        # check if the daemon binary exists
        return path.isfile(f"{FRR_DAEMONPATH}{daemon}")

    return False



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from topology sub-package"""

import unittest
import subprocess

from nest.topology import Node, connect, Switch, Router
from nest.clean_up import delete_namespaces
from nest.topology_map import TopologyMap
from nest.topology.interface import Interface
import nest.config as config

# pylint: disable=missing-docstring
# pylint: disable=invalid-name
class TestTopology(unittest.TestCase):
    def setUp(self):
        self.n0 = Node("n0")
        self.n1 = Node("n1")

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def test_p2p(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address("10.0.0.1/24")
        n1_n0.set_address("10.0.0.2/24")

        status = self.n0.ping("10.0.0.2")

        self.assertTrue(status)

    def test_p2p_ipv6(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        self.n0.disable_ip_dad()
        self.n1.disable_ip_dad()

        n0_n1.set_address("2001:1:1:1443::411/122")
        n1_n0.set_address("2001:1:1:1443::412/122")

        status = self.n0.ping(n1_n0.address)

        self.assertTrue(status)

    def test_prp(self):
        # pylint: disable=invalid-name
        r = Node("r")
        r.enable_ip_forwarding()

        (n0_r, r_n0) = connect(self.n0, r)
        (r_n1, n1_r) = connect(r, self.n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping("10.1.2.1")

        self.assertTrue(status)

    def test_prp_ipv6(self):
        # pylint: disable=invalid-name
        r = Node("r")
        r.enable_ip_forwarding(ipv6=True)

        (n0_r, r_n0) = connect(self.n0, r)
        (r_n1, n1_r) = connect(r, self.n1)

        self.n0.disable_ip_dad()
        self.n1.disable_ip_dad()

        n0_r.set_address("2001:0001:0001:1443::0411/122")
        r_n0.set_address("2001:0001:0001:1443::0412/122")
        r_n1.set_address("2001:0001:0001:1444::0412/122")
        n1_r.set_address("2001:0001:0001:1444::0411/122")

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping(n1_r.address)

        self.assertTrue(status)

    def test_simple_lan(self):
        # pylint: disable=too-many-locals
        n0 = Node("n0")
        n1 = Node("n1")
        n2 = Node("n2")
        n3 = Node("n3")
        s0 = Switch("s0")

        (n0_s0, _) = connect(n0, s0)
        (n1_s0, _) = connect(n1, s0)
        (n2_s0, _) = connect(n2, s0)
        (n3_s0, _) = connect(n3, s0)

        nodes = [n0, n1, n2, n3]
        interface = [n0_s0, n1_s0, n2_s0, n3_s0]

        n0_s0.set_address("10.0.0.1/24")
        n1_s0.set_address("10.0.0.2/24")
        n2_s0.set_address("10.0.0.3/24")
        n3_s0.set_address("10.0.0.4/24")

        for x in nodes:
            for y in interface:
                status = x.ping(y.address, packets=1)
                if not status:
                    break

        self.assertTrue(status)

    def test_dumbbell_lan(self):
        # pylint: disable=too-many-locals
        n0 = Node("n0")
        n1 = Node("n1")
        n2 = Node("n2")
        n3 = Node("n3")
        s0 = Switch("s0")
        s1 = Switch("s1")

        (n0_s0, _) = connect(n0, s0)
        (n1_s0, _) = connect(n1, s0)
        (n2_s1, _) = connect(n2, s1)
        (n3_s1, _) = connect(n3, s1)

        connect(s0, s1)

        nodes = [n0, n1, n2, n3]
        interface = [n0_s0, n1_s0, n2_s1, n3_s1]

        n0_s0.set_address("10.0.0.1/24")
        n1_s0.set_address("10.0.0.2/24")
        n2_s1.set_address("10.0.0.3/24")
        n3_s1.set_address("10.0.0.4/24")

        for x in nodes:
            for y in interface:
                status = x.ping(y.address, packets=1)
                if not status:
                    break

        self.assertTrue(status)

    def test_tcp_param(self):
        self.n0.configure_tcp_param("ecn", "1")
        ecn = self.n0.read_tcp_param("ecn")
        self.assertEqual(ecn, "1")

    def test_invalid_interface_name(self):
        # Disable topology map
        config.set_value("assign_random_names", False)

        # Valid interface name
        Interface("namewith15chars")

        # Invalid interface name
        with self.assertRaises(ValueError) as cm:
            Interface("looonginvalidname")
        err = cm.exception
        self.assertEqual(
            str(err),
            "Device name looonginvalidname is too long. Device names "
            "should not exceed 15 characters",
        )

        # Enable topology map
        config.set_value("assign_random_names", True)

    def test_invalid_veth_name(self):
        # Disable topology map
        config.set_value("assign_random_names", False)

        node0 = Node("longname0")
        node1 = Node("longname1")

        with self.assertRaises(ValueError) as cm:
            connect(node0, node1)
        err = cm.exception
        self.assertEqual(
            str(err),
            "Auto-generated device name longname0-longname1-0 is "
            "too long. The length of name should not exceed 15 characters.",
        )

        # Enable topology map
        config.set_value("assign_random_names", True)

    def test_valid_veth_name(self):
        node0 = Node("node0")
        node1 = Node("node1")

        eth0, eth1 = connect(node0, node1)

        self.assertEqual(
            eth0.name, "node0-node1-0", "eth0 has unexpected autogenerated name"
        )
        self.assertEqual(
            eth1.name, "node1-node0-0", "eth1 has unexpected autogenerated name"
        )

        eth2, eth3 = connect(node0, node1)

        self.assertEqual(
            eth2.name, "node0-node1-1", "eth2 has unexpected autogenerated name"
        )
        self.assertEqual(
            eth3.name, "node1-node0-1", "eth3 has unexpected autogenerated name"
        )

        eth4, eth5 = connect(node0, node1, "eth4", "eth5")

        self.assertEqual(eth4.name, "eth4", "eth4 has unexpected user-given name")
        self.assertEqual(eth5.name, "eth5", "eth5 has unexpected user-given name")

    def test_invalid_ifb_interface_name(self):
        # Disable topology map
        config.set_value("assign_random_names", False)

        node0 = Node("node0")
        node1 = Node("node1")
        eth0, _ = connect(node0, node1)

        with self.assertRaises(ValueError) as cm:
            eth0.set_attributes("10mbit", "10ms", "codel")
        err = cm.exception
        self.assertEqual(
            str(err),
            "Device name ifb-node0-node1-0 is too "
            "long. Device names should not exceed 15 characters",
        )

        # Enable topology map
        config.set_value("assign_random_names", True)

    def test_run_inside_node(self):
        (n0_n1, n1_n0) = connect(self.n0, self.n1)

        n0_n1.set_address("10.0.0.1/24")
        n1_n0.set_address("10.0.0.2/24")

        # Run ping from self.n0 to self.n1
        with self.n0:
            command = f"ping -c 1 {n1_n0.address.get_addr(with_subnet=False)}"
            with subprocess.Popen(
                command.split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ) as proc:

                (stdout, _) = proc.communicate()

        self.assertEqual(stdout[:4], b"PING", "Invalid ping output")

    def test_prp_router_api(self):
        # pylint: disable=invalid-name
        r = Router("r")

        (n0_r, r_n0) = connect(self.n0, r)
        (r_n1, n1_r) = connect(r, self.n1)

        n0_r.set_address("10.1.1.1/24")
        r_n0.set_address("10.1.1.2/24")
        r_n1.set_address("10.1.2.2/24")
        n1_r.set_address("10.1.2.1/24")

        self.n0.add_route("DEFAULT", n0_r)
        self.n1.add_route("DEFAULT", n1_r)

        status = self.n0.ping("10.1.2.1")

        self.assertTrue(status)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Sysctl commands"""
import logging
from .exec import exec_subprocess

logger = logging.getLogger(__name__)


def en_ip_forwarding(ns_name, ipv4=True, ipv6=True):
    """
    Enables ip forwarding in a namespace. Used for routers

    Parameters
    ----------
    ns_name : str
        namespace name
    ipv4 : bool
        if `True`, enables ipv4 ip forwarding
    ipv6 : bool
        if `True`, enables ipv6 ip forwarding
    """
    if ipv6:
        exec_subprocess(
            f"ip netns exec {ns_name} sysctl -w net.ipv6.conf.all.forwarding=1"
        )
    if ipv4:
        exec_subprocess(f"ip netns exec {ns_name} sysctl -w net.ipv4.ip_forward=1")


def disable_dad(ns_name, int_name):
    """
    Disables DAD at nodes for IPv6 Addressing

    Parameters
    ----------
    ns_name : str
        namespace name
    int_name : str
        interface name
    """
    exec_subprocess(
        f"ip netns exec {ns_name} sysctl -w net.ipv6.conf.{int_name}.accept_dad=0"
    )


def configure_kernel_param(ns_name, prefix, param, value):
    """
    Configure kernel parameters using sysctl

    Parameters
    ----------
    ns_name : str
        name of the namespace
    prefix : str
        path for the sysctl command
    param : str
        kernel parameter to be configured
    value : str
        value of the parameter
    """
    exec_subprocess(f"ip netns exec {ns_name} sysctl -q -w {prefix}{param}={value}")


def read_kernel_param(ns_name, prefix, param):
    """
    Read kernel parameters using sysctl

    Parameters
    ----------
    ns_name : str
        name of the namespace
    prefix : str
        path for the sysctl command
    param : str
        kernel parameter to be read

    Returns
    -------
    str
        value of the `param`
    """
    value = exec_subprocess(
        f"ip netns exec {ns_name} sysctl -n {prefix}{param}", output=True
    )
    return value.rstrip("\n")


def set_mpls_max_label_node(ns_name, max_num_labels):
    """
    Sets the maximum mpls label for the namespace

    Parameters
    ----------
    ns_name: str
    max_num_labels: int
    """
    exec_subprocess(
        f"ip netns exec {ns_name} sysctl -w net.mpls.platform_labels={max_num_labels}"
    )
    try:
        max_label = int(
            exec_subprocess(
                f"ip netns exec {ns_name} sysctl -n net/mpls/platform_labels",
                output=True,
            )
        )
        if max_label != max_num_labels:
            logger.error(
                "platform_labels for node %s wasn't set to %s!",
                str(ns_name),
                str(max_num_labels),
            )
            logger.error("Ensure mpls kernel modules are loaded")
    except ValueError:
        logger.error(
            "Couldn't set platform_labels. Ensure mpls kernel modules are loaded"
        )


def enable_mpls_interface(ns_name, dev_name):
    """
    Enable mpls input to interface through sysctl.

    Parameters
    ----------
    ns_name: str
    dev_name: str
    """
    exec_subprocess(
        f"ip netns exec {ns_name} sysctl -w net.mpls.conf.{dev_name}.input=1",
        output=True,
    )
    try:
        input_val = int(
            exec_subprocess(
                f"ip netns exec {ns_name} sysctl -n net.mpls.conf.{dev_name}.input",
                output=True,
            )
        )
        if input_val == 0:
            logger.error("Couldn't enable mpls input for interface %s!!", dev_name)
    except ValueError:
        logger.error(
            "Couldn't enable mpls input. Ensure mpls kernel modules are loaded"
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test APIs from routing sub-package"""

import unittest
from glob import glob
from nest import config
from nest.topology_map import TopologyMap
from nest.topology import Node, connect
from nest.routing.routing_helper import RoutingHelper
from nest.clean_up import delete_namespaces

# pylint: disable=missing-docstring


class TestQuagga(unittest.TestCase):

    # pylint: disable=invalid-name
    def setUp(self):
        self.n0 = Node("n0")
        self.n1 = Node("n1")
        self.r0 = Node("r0")
        self.r1 = Node("r1")
        self.r0.enable_ip_forwarding()
        self.r1.enable_ip_forwarding()

        ### Create interfaces and connect nodes and routers ###

        (eth_p1r1, eth_r1p1) = connect(self.n0, self.r0, "eth-n1r1-0", "eth-r1n1-0")
        (eth_r1r2, eth_r2r1) = connect(self.r0, self.r1, "eth-r1r2-0", "eth-r2r1-0")
        (eth_r2p2, eth_p2r2) = connect(self.r1, self.n1, "eth-r2n2-0", "eth-n2r2-0")

        ### Assign addresses to interfaces ###

        eth_p1r1.set_address("10.0.1.1/24")
        eth_r1p1.set_address("10.0.1.2/24")

        eth_r1r2.set_address("10.0.2.2/24")
        eth_r2r1.set_address("10.0.2.3/24")

        eth_r2p2.set_address("10.0.3.3/24")
        eth_p2r2.set_address("10.0.3.4/24")

        config.set_value("routing_suite", "quagga")  # Use quagga

    def tearDown(self):
        delete_namespaces()
        TopologyMap.delete_all_mapping()

    def test_routing_helper(self):

        RoutingHelper("rip").populate_routing_tables()

        status = self.n0.ping("10.0.3.4", verbose=False)
        self.assertTrue(status)

        status = self.n1.ping("10.0.1.1", verbose=False)
        self.assertTrue(status)

    def test_ospf(self):
        RoutingHelper("ospf").populate_routing_tables()

        status = self.n0.ping("10.0.3.4", verbose=False)
        self.assertTrue(status)

        status = self.n1.ping("10.0.1.1", verbose=False)
        self.assertTrue(status)

    def test_isis(self):
        RoutingHelper("isis").populate_routing_tables()

        status = self.n0.ping("10.0.3.4", verbose=False)
        self.assertTrue(status)

        status = self.n1.ping("10.0.1.1", verbose=False)
        self.assertTrue(status)

    def test_logs(self):
        config.set_value("routing_logs", True)

        RoutingHelper("rip").populate_routing_tables()

        self.assertTrue(len(glob(f"{config.get_value('routing_suite')}-logs_*")) > 0)

        config.set_value("routing_logs", False)

    def test_custom_node_routers(self):
        RoutingHelper(
            "rip", [self.n0, self.n1], [self.r0, self.r1]
        ).populate_routing_tables()

        status = self.n0.ping("10.0.3.4", verbose=False)
        self.assertTrue(status)

        status = self.n1.ping("10.0.1.1", verbose=False)
        self.assertTrue(status)

        with self.assertRaises(TypeError):
            RoutingHelper("rip", self.n1, self.r1).populate_routing_tables()

        with self.assertRaises(ValueError):
            RoutingHelper("rip", ["n1"], ["r1"]).populate_routing_tables()


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""ip route commands (mpls family)"""

from .exec import exec_subprocess


def add_mpls_route_push(host_name, dest_ip, next_hop_ip, label):
    """
    Adds a route in routing table of host.

    Parameters
    ----------
    host_name : str
        name of the host namespace
    dest_ip : str
        the destination ip for the route
    next_hop_ip : str
        IP of the very next interface
    label : str
        the label that is to be pushed onto the packet
    """
    exec_subprocess(
        f"ip netns exec {host_name} ip route add {dest_ip}"
        f" encap mpls {label} via inet {next_hop_ip}"
    )


def add_mpls_route_switch(host_name, incoming_label, next_hop_ip, outgoing_label):
    """
    Adds a route in routing table of host.

    Parameters
    ----------
    host_name : str
        name of the host namespace
    incoming_label : str
        the label that is on the packet when it arrives at the node
    next_hop_ip : str
        IP of the very next interface
    outgoing_label : str
        the label that is to be pushed onto the packet
    """
    exec_subprocess(
        f"ip netns exec {host_name} ip -f mpls route add {incoming_label}"
        f" as {outgoing_label} via inet {next_hop_ip}"
    )


def add_mpls_route_pop(host_name, incoming_label, next_hop_ip):
    """
    Adds a route in routing table of host.

    Parameters
    ----------
    host_name : str
        name of the host namespace
    next_hop_ip : str
        IP of the very next interface
    incoming_label : str
        the label that is on the packet when it arrives at the node
    """
    exec_subprocess(
        f"ip netns exec {host_name} ip -f mpls route add {incoming_label}"
        f" via inet {next_hop_ip}"
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
NeST: Network Stack Tester
==========================

This is the entry point of NeST package.

Following actions are performed as part of setup, when
nest is imported:
1. Check if nest is run with root privileges
2. Store SUDO user and group id information
"""

import logging
import os
import sys
import signal

from nest import clean_up
from .logging_helper import add_logging_level, get_trace_filehandler
from .user import User
from . import config

# Set high logging level so that logs aren't printed
# if user is not running as root
nest_logger = logging.getLogger(__name__)
nest_logger.setLevel(logging.CRITICAL)

if os.geteuid() != 0:
    print("nest: python package requires root access", file=sys.stderr)
    sys.exit(1)

# Load default config values
config.import_default_config()

# Store user information for later use
if all(key in os.environ for key in ("SUDO_UID", "SUDO_GID")):
    user_id = int(os.environ["SUDO_UID"])
    group_id = int(os.environ["SUDO_GID"])
    User(user_id, group_id)

# Set up logging
log_level = config.get_value("log_level")

# Logging level TRACE is used to output all the commands executed by engine to a file
add_logging_level("TRACE", logging.DEBUG - 1, "trace")

nest_logger.setLevel(log_level)
ch = logging.StreamHandler()  # Logger output will be output to stderr
ch.setLevel(log_level)
formatter = logging.Formatter("[%(levelname)s] : %(message)s")
ch.setFormatter(formatter)

# pylint: disable=no-member
ch.addFilter(
    lambda record: record.levelno != logging.TRACE
)  # To avoid engine commands to be printed to stdout
nest_logger.addHandler(ch)

if log_level == "TRACE":
    nest_logger.addHandler(get_trace_filehandler())

# On recieving Termination signal, execute the given function
signal.signal(signal.SIGTERM, clean_up.delete_namespaces)

# Load custom config values
config.search_config_files()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Helper class for routing
"""

import time
import logging
import importlib
from os import mkdir, kill, path, listdir
import atexit
from shutil import rmtree, chown
from signal import SIGTERM
from typing import List
from nest.exceptions import RequiredDependencyNotFound
from nest.input_validator.input_validator import input_validator
from nest.topology.id_generator import IdGen
from nest.routing.zebra import Zebra
from nest.routing.ldp import Ldp
from nest.topology_map import TopologyMap
from nest.user import User
from nest import config
from nest.topology import Node

logger = logging.getLogger(__name__)

# pylint:disable=too-few-public-methods
# pylint:disable=too-many-instance-attributes


class RoutingHelper:
    """
    Handles basic routing requirements for the topology.
    Either inherit this class or use `Zebra` and other protocols
    for better customizations

    Attributes
    ----------
    _module_map : dict
        map between protocol string and its module and class
    protocol : str
        routing protocol(one of ['rip', 'ospf', 'isis'])
    routers : List[Node]
        routers in the topology
    hosts : List[Node]
        hosts in the topology
    conf_dir : str
        path for config directory of daemons
    protocol_class : ABCMeta
        Protocol class which will later be instantiated
    """

    _module_map = {
        "ospf": ["nest.routing.ospf", "Ospf"],
        "rip": ["nest.routing.rip", "Rip"],
        "isis": ["nest.routing.isis", "Isis"],
    }

    @input_validator
    def __init__(
        self,
        protocol: str,
        hosts: List[Node] = None,
        routers: List[Node] = None,
        ldp_routers: List[Node] = None,
    ):
        """
        Constructor for RoutingHelper.
        The dynamic routing daemons will be run only on nodes with more than
        one interface. Specify `hosts` & `routers` parameters to override this.

        Parameters
        ----------
        protocol: str
            routing protocol to be run. One of [ospf, rip, isis]
        hosts: List[Node]
            List of hosts in the network. If `None`, considers the entire topology.
            Use this if your topology has disjoint networks
        routers: List[Node]
            List of routers in the network. If `None`, considers the entire topology.
            Use this if your topology has disjoint networks
        ldp_routers: List[Node]
            List of Routers which are to be used with mpls.
            Only enables ldp discovery on interfaces with mpls enabled
        """
        if protocol == "static":
            raise NotImplementedError(
                "Static routing is yet to be implemented. Use rip, ospf or isis"
            )
        if protocol not in ["rip", "ospf", "isis"]:
            raise ValueError(
                f"Supported routing protocols are rip, ospf and isis, "
                f"but got protocol {protocol}"
            )
        self.protocol = protocol

        # Validate hosts, routers and ldp_routers
        self._is_node_list("hosts", hosts)
        self._is_node_list("routers", routers)
        self._is_node_list("ldp_routers", ldp_routers)

        self.hosts = []
        self.routers = []
        if routers is None and hosts is None:
            all_nodes = TopologyMap.get_hosts() + TopologyMap.get_routers()
            for node in all_nodes:
                num_interfaces = len(node.interfaces)
                if num_interfaces == 1:
                    self.hosts.append(node)
                elif num_interfaces > 1:
                    self.routers.append(node)
        else:
            self.hosts = hosts
            self.routers = routers

        self.ldp_routers = ldp_routers if ldp_routers is not None else []
        self.conf_dir = None
        self.log_dir = None
        module_str, class_str = RoutingHelper._module_map[self.protocol]
        module = importlib.import_module(module_str)
        self.protocol_class = getattr(module, class_str)
        self.zebra_list = []
        self.protocol_list = []
        self.ldp_list = []

        atexit.register(self._clean_up)

    def populate_routing_tables(self):
        """
        Populate routing table using `self.protocol`
        """
        self._setup_default_routes()

        if self.protocol == "static":
            pass  # TODO: add static routing
        else:
            try:
                self._run_dyn_routing()
            except RequiredDependencyNotFound:
                return

    def _create_directory(self, dir_path):
        """
        Creates a quagga/frr owned directory at `dir_path`

        Parmeters
        ---------
        dir_path: path of the directory to be created

        """
        mkdir(dir_path)
        chown(dir_path, user=config.get_value("routing_suite"))

    def _create_conf_directory(self):
        """
        Creates a directory for holding routing related config
        and pid files.
        Override this to create directory at a location other than /tmp

        Returns
        -------
        str:
            path of the created directory
        """
        salt = config.get_value("routing_suite") + str(time.clock_gettime(0))
        dir_path = f"/tmp/{salt}-configs_{IdGen.topology_id}"
        self._create_directory(dir_path)
        return dir_path

    def _create_log_directory(self):
        """
        Creates a directory for holding routing log files.

        Returns
        -------
        str:
            path of the created directory
        """
        timestamp = time.strftime("%d-%m-%Y-%H:%M:%S")
        log_path = f"{config.get_value('routing_suite')}-logs_{timestamp}"
        self._create_directory(log_path)
        return log_path

    def _setup_default_routes(self):
        """
        Setup default routes in hosts
        """
        for host in self.hosts:
            host.add_route("DEFAULT", host.interfaces[0])

    def _run_dyn_routing(self):
        """
        Run zebra and `self.protocol`
        """
        logger.info("Running zebra and %s on routers", self.protocol)
        self.conf_dir = self._create_conf_directory()
        if config.get_value("routing_logs"):
            self.log_dir = self._create_log_directory()

        for router in self.routers:
            self._run_zebra(router)
            self._run_routing_protocol(router)
            if router in self.ldp_routers:
                self._run_ldp(router)
        self._check_for_convergence()

    def _run_zebra(self, router):
        """
        Create required config file and run zebra
        """
        zebra = Zebra(router.id, router.interfaces, self.conf_dir, log_dir=self.log_dir)
        zebra.create_basic_config()
        zebra.run()
        self.zebra_list.append(zebra)

    def _run_routing_protocol(self, router):
        """
        Create required config file and run `self.protocol`
        """
        protocol = self.protocol_class(
            router.id, router.interfaces, self.conf_dir, log_dir=self.log_dir
        )
        protocol.create_basic_config()
        protocol.run()
        self.protocol_list.append(protocol)

    def _run_ldp(self, router):
        """
        Create required config file and run ldp
        """
        mpls_interfaces = []
        for interface in router.interfaces:
            if interface.is_mpls_enabled():
                mpls_interfaces.append(interface)
        if len(mpls_interfaces) == 0:
            raise Exception("MPLS isn't enabled in any interface!")
        ldp = Ldp(router.id, mpls_interfaces, self.conf_dir, log_dir=self.log_dir)
        ldp.create_basic_config()
        ldp.run()
        self.ldp_list.append(ldp)

    def _check_for_convergence(self):
        """
        Wait for the routing protocol to converge.
        Override this for custom convergence check
        """
        logger.info("Waiting for %s to converge", self.protocol)
        interval = 2
        converged = False
        # Ping between hosts until convergence
        while not converged:
            time.sleep(interval)
            converged = True
            for i in range(len(self.hosts)):
                for j in range(i + 1, len(self.hosts)):
                    if not self.hosts[i].ping(
                        self.hosts[j].interfaces[0].address.get_addr(), verbose=False
                    ):
                        converged = False
                        break
                if not converged:
                    break

        logger.info("Routing completed")

    def _is_node_list(self, arg_name, node_list):
        """
        Checks if `node_list` is a list of Nodes.
        """
        if node_list is None:
            return
        for node in node_list:
            if isinstance(node, Node) is False:
                raise ValueError(
                    f"Some items in argument '{arg_name}' of method '{self.__init__.__qualname__}' "
                    f"are not derived from type '{Node.__name__}'"
                )

    def _clean_up(self):
        """
        Terminates routing daemons and deletes config files
        """
        # To preserve the routes, the daemons shouldn't be stopped
        # Frrouting doesn't seem to have an option to not flush the routes installed

        # Stop ldp processes
        for ldp in self.ldp_list:
            if path.isfile(ldp.pid_file):
                with open(ldp.pid_file, "r") as pid_file:
                    pid = int(pid_file.read())
                    try:
                        kill(pid, SIGTERM)
                    except ProcessLookupError:
                        pass

        # Stop zebra processes
        for zebra in self.zebra_list:
            if path.isfile(zebra.pid_file):
                with open(zebra.pid_file, "r") as pid_file:
                    pid = int(pid_file.read())
                    try:
                        kill(pid, SIGTERM)
                    except ProcessLookupError:
                        pass

        # Stop protocol processes
        for protocol in self.protocol_list:
            if path.isfile(protocol.pid_file):
                with open(protocol.pid_file, "r") as pid_file:
                    pid = int(pid_file.read())
                    try:
                        kill(pid, SIGTERM)
                    except ProcessLookupError:
                        pass

        # Delete config directory
        if path.isdir(self.conf_dir):
            rmtree(self.conf_dir)

        # Change ownership of log files to current user
        if self.log_dir is not None and path.isdir(self.log_dir):
            chown(self.log_dir, user=User.user_id, group=User.group_id)
            for file in listdir(self.log_dir):
                chown(
                    path.join(self.log_dir, file),
                    user=User.user_id,
                    group=User.group_id,
                )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Ping command"""

from .exec import exec_exp_commands, exec_subprocess, exec_subprocess_with_live_output


def ping(ns_name, dest_addr, packets=1, ipv6=False, live_output=True):
    """
    Send a ping packet from ns_name to dest_addr
    if possible

    Parameters
    ----------
    ns_name : str
        namespace name
    dest_addr : str
        address to ping to
    packets : int
        Number of ping packets sent (default: 1)
    live_output : bool
        Show live output of ping packets

    Returns
    -------
    bool
        success of ping
    """
    if live_output:
        if ipv6:
            status = exec_subprocess_with_live_output(
                f"ip netns exec {ns_name} ping -6 -c {packets} {dest_addr}"
            )
        else:
            status = exec_subprocess_with_live_output(
                f"ip netns exec {ns_name} ping -c {packets} {dest_addr}"
            )
    else:
        if ipv6:
            status = exec_subprocess(
                f"ip netns exec {ns_name} ping -6 -c {packets} {dest_addr}"
            )
        else:
            status = exec_subprocess(
                f"ip netns exec {ns_name} ping -c {packets} {dest_addr}"
            )
    return status == 0


# pylint: disable=too-many-arguments
def run_exp_ping(ns_id, destination_ip, run_time, ipv6, out, err):
    """
    Run ping to extract stats

    Parameters
    ----------
    ns_id : str
        network namespace to run netperf from
    destination_ip : str
        IP address of the destination namespace
    run_time : num
        total time to run netperf for
    ipv6 : bool
        determines if destination_ip is ipv4/ipv6
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    if ipv6:
        return exec_exp_commands(
            f"ip netns exec {ns_id} ping -6 {destination_ip} -w {run_time} -D \
            -i 0.2",
            stdout=out,
            stderr=err,
        )

    return exec_exp_commands(
        f"ip netns exec {ns_id} ping {destination_ip} -w {run_time} -D \
            -i 0.2",
        stdout=out,
        stderr=err,
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

""" APIs to process config files """

import os
import json
import logging
from nest.logging_helper import update_nest_logger
from nest.input_validator import input_validator
from nest.exception import NestBaseException

logger = logging.getLogger(__name__)

__DEFAULT_VALUE = {}
__DEFAULT_PATH = [
    "/etc/nest-config.json",
    os.path.expanduser("~") + "/.nest-config.json",
    os.getcwd() + "/nest-config.json",
]


def import_default_config():
    """
    Reads `config.json` file from necessary specified file locations and converts it into a string

    Returns
    -------
    dict
        The information in the default config file as a python dictionary
    """
    # Having it as a private, global variable prevents users from directly accessing it
    # pylint: disable=global-statement

    global __DEFAULT_VALUE
    with open(
        os.path.realpath(os.path.dirname(__file__)) + "/config.json", "r"
    ) as json_file:
        data = json.load(json_file)

    __DEFAULT_VALUE = data


@input_validator
def set_value(parameter: str, value):
    """
    Changes the default values only for that program

    Attributes
    ----------
    parameter: str
        The parameter's value which has to be changed
    value: str
        The value to which the parameter has to be changed to
    """
    if parameter in __DEFAULT_VALUE:
        if isinstance(value, type(__DEFAULT_VALUE[parameter])):
            __DEFAULT_VALUE[parameter] = value
        else:
            raise ValueError(
                f"The value of config '{parameter}' should be of type "
                f"{type(__DEFAULT_VALUE[parameter])}, but got value "
                f"'{value}' of type {type(value)}"
            )
    else:
        raise ValueError(f"The given config parameter {parameter} does not exist")

    _post_set_value(parameter, value)


def _post_set_value(parameter, value):
    """
    Called after setting a config value. Used for executing
    parameter specific code.

    Attributes
    ----------
    parameter: str
        The parameter's value which has to be changed
    value: str
        The value to which the parameter has to be changed to
    """

    if parameter == "log_level":
        update_nest_logger(value)


def get_value(parameter):
    """
    Returns the default value of the parameter

    Attributes
    ----------
    parameter: str
        The parameter whose value is to be returned

    Returns
    -------
    str
        The value of the parameter
    """
    if parameter in __DEFAULT_VALUE:
        return __DEFAULT_VALUE[parameter]
    # If it belonged to none of them
    logger.error("The given parameter %s does not exist", parameter)
    return None


def import_custom_config(path: str):
    """
    Overwrites the current value of config parameters with the values
    from the JSON in the given path

    Attributes
    ----------
    path: str
        The path in which the json file
    """
    with open(path, "r") as json_file:
        data = json.load(json_file)
    for parameter in data:
        set_value(parameter, data[parameter])


def search_config_files():
    """
    Searches a few predifined locations for a config files
    """

    # The ones called later, might overwrite the previous calls
    for config_path in __DEFAULT_PATH:
        try:
            if os.path.isfile(os.path.abspath(config_path)):
                import_custom_config(os.path.abspath(config_path))
        except ValueError as error:
            raise NestConfigInitializationError(
                f"An error occurred when initializing NeST config from {config_path}. "
                f"Please see the previous exception for more details about the error."
            ) from error


class NestConfigInitializationError(NestBaseException):
    """
    Exception related to error in Initializaton of NeST
    config from config file.
    """



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Stores user information"""

# pylint: disable=too-few-public-methods
class User:
    """
    User info (user id and group id) stored here.

    Attributes
    ----------
    user_id : int
        User id of user
    group_id : int
        Group id of user

    """

    user_id = -1
    group_id = -1

    def __init__(self, user_id, group_id):
        """
        Initialize user info.

        Parameters
        ----------
        user_id : int
        group_id : int

        """
        User.user_id = user_id
        User.group_id = group_id



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""IP link commands"""
import logging
from .exec import exec_subprocess

logger = logging.getLogger(__name__)


def create_switch(ns_name, s_name):
    """
    Create switch if it doesn't already exist.
    Switch is contained inside a namespace.

    Parameters
    ----------
    s_name : str
        switch name
    ns_name : str
        namespace name
    """
    exec_subprocess(f"ip netns exec {ns_name} ip link add {s_name} type bridge")


def delete_switch(ns_name, s_name):
    """
    Delete switch inside a namespace, if it exists.

    Parameters
    ----------
    s_name : str
        switch name
    ns_name : str
        namespace name
    """
    exec_subprocess(f"ip netns exec {ns_name} ip link del {s_name}")


def set_switch_mode(s_name, mode):
    """
    Set switch mode

    Parameters
    ----------
    s_name : str
        switch name
    mode : str
        'up' or 'down'
    """
    exec_subprocess(f"ip netns exec {s_name} ip link set dev {s_name} {mode}")


def add_int_to_switch(s_name, dev_name):
    """
    Add interface to a switch

    Parameters
    ----------
    s_name : str
        switch name
    dev_name : str
        interface name
    """
    exec_subprocess(f"ip netns exec {s_name} ip link set {dev_name} master {s_name}")


def create_veth(dev_name1, dev_name2):
    """
    Create a veth pair with endpoint interfaces `dev_name1`
    and `dev_name2`

    Parameters
    ----------
    dev_name1 : str
    dev_name2 : str
    """
    exec_subprocess(f"ip link add {dev_name1} type veth peer name {dev_name2}")


def create_ifb(dev_name):
    """
    Create a IFB

    Parameters
    ----------
    dev_name : str
        interface names
    """

    exec_subprocess(f"ip link add {dev_name} type ifb")


def add_int_to_ns(ns_name, dev_name):
    """
    Add interface to a namespace

    Parameters
    ----------
    ns_name : str
        namespace name
    dev_name : str
        interface name
    """
    exec_subprocess(f"ip link set {dev_name} netns {ns_name}")


def set_int_up(ns_name, dev_name):
    """
    Set interface mode to up

    Parameters
    ----------
    ns_name : str
        namespace name
    dev_name : str
        interface name
    """
    exec_subprocess(f"ip netns exec {ns_name} ip link set dev {dev_name} up")


def setup_veth(ns_name1, ns_name2, dev_name1, dev_name2):
    """
    Sets up veth connection between interfaces. The interfaces are
    set up as well.

    The connections are made between `dev_name1` in `ns_name1` and
    `dev_name2` in `ns_name2`

    Parameters
    ----------
    ns_name1 : str
    dev_name1 : str
    dev_name2 : str
    ns_name2 : str
    """
    create_veth(dev_name1, dev_name2)
    add_int_to_ns(ns_name1, dev_name1)
    add_int_to_ns(ns_name2, dev_name2)
    set_int_up(ns_name1, dev_name1)
    set_int_up(ns_name2, dev_name2)


def setup_ifb(ns_name, dev_name):
    """
    Sets up an IFB device. The device is setup as well.

    Parameters
    ----------
    ns_name : str
        namespace name
    dev_name : str
        name of IFB
    """

    create_ifb(dev_name)
    add_int_to_ns(ns_name, dev_name)
    set_int_up(ns_name, dev_name)


def set_interface_mode(ns_name, dev_name, mode):
    """
    Set interface mode up or down

    Parameters
    ----------
    ns_name : str
    dev_name : str
    mode : str
        'up' or 'down'
    """
    exec_subprocess(f"ip netns exec {ns_name} ip link set dev {dev_name} {mode}")


def set_mtu_interface(ns_name, dev_name, mtu_value):
    """
    Set the MTU for the interface

    Parameters
    ----------
    ns_name : str
    dev_name : str
    mtu_value : int
    """
    exec_subprocess(
        f"ip netns exec {ns_name} ip link set dev {dev_name} mtu {mtu_value}"
    )
    # Verify if the mtu is set
    mtu = int(
        exec_subprocess(
            f" ip netns exec {ns_name} cat /sys/class/net/{dev_name}/mtu", output=True
        )
    )
    if mtu != mtu_value:
        logger.error(
            "MTU of interface %s wasn't set to %s!", str(dev_name), str(mtu_value)
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""
Runs netperf commands to setup TCP flow
and collect throughput data
"""

import re
import copy
import logging
from time import sleep
from functools import partial
from nest.experiment.interrupts import handle_keyboard_interrupt
from ..results import NetperfResults
from .runnerbase import Runner
from ...topology_map import TopologyMap
from ...engine.netperf import run_netperf, run_netserver

logger = logging.getLogger(__name__)


class NetperfRunner(Runner):
    """
    Runs netperf command and parses statistics from its output

    Attributes
    ----------
    default_netperf_options : dict
        default options to run netperf command with
    netperf_tcp_options : dict
        tcp related netperf options
    netperf_udp_options : dict
        udp related netperf options
    ns_id : str
        network namespace to run netperf from
    destination_ip : str
        ip address of the destination namespace
    start_time : num
        time at which netperf is to run
    run_time : num
        total time to run netperf for
    """

    tcp_output_options = [
        "THROUGHPUT",
        "LOCAL_CONG_CONTROL",
        "REMOTE_CONG_CONTROL",
        "TRANSPORT_MSS",
        "LOCAL_SOCKET_TOS",
        "REMOTE_SOCKET_TOS",
    ]

    # fmt: off

    default_netperf_options = {
        "banner": "-P 0",                   # Disable test banner
        "testname": "-t TCP_STREAM",        # Test type (NOTE: TCP_STREAM only for now)
        "fill_file": "-F /dev/urandom",     # File to transmit (NOTE: Inspired from flent)
        "testlen": "-l 10",                 # Length of test (NOTE: Default 10s)
        "interval": "-D -0.2",              # Generate interim results every INTERVAL secs
        "debug": "-d",                      # Enable debug mode
    }

    netperf_tcp_options = {
        "cong_algo": "-K cubic",            # Congestion algorithm
        "stats": "-k THROUGHPUT",           # Stats required
    }

    netperf_udp_options = {
        "routing": "-R 1",                  # Enable routing
        "stats": "-k THROUGHPUT",           # Stats required
    }

    # fmt: on

    # pylint: disable=too-many-arguments
    def __init__(self, ns_id, destination_ip, start_time, run_time, dst_ns, **kwargs):
        """
        Constructor to initialize netperf runner

        Parameters
        ----------
        ns_id : str
            network namespace to run netperf from
        destination_ip : str
            ip address of the destination namespace
        start_time : num
            time at which netperf is to run
        run_time : num
            total time to run netperf for
        dst_ns : str
            network namespace to run netperf server in
        **kwargs
            netperf options to override
        """
        self.options = copy.deepcopy(kwargs)
        super().__init__(ns_id, start_time, run_time, destination_ip, dst_ns)

    # Should this be placed somewhere else?
    @staticmethod
    def run_netserver(ns_id):
        """
        Run netserver in `ns_id`

        Parameters
        ----------
        ns_id : str
            namespace to run netserver on
        """
        return_code = run_netserver(ns_id)
        if return_code != 0:
            ns_name = TopologyMap.get_namespace(ns_id)["name"]
            logger.error("Error running netserver at %s.", ns_name)

    def run(self):
        """
        Runs netperf at t=`self.start_time`
        """
        netperf_options = copy.copy(NetperfRunner.default_netperf_options)
        test_options = None

        # Change the default run time
        netperf_options["testlen"] = f"-l {self.run_time}"

        # Set test
        netperf_options["testname"] = f"-t {self.options['testname']}"

        if netperf_options["testname"] == "-t TCP_STREAM":
            test_options = copy.copy(NetperfRunner.netperf_tcp_options)
            test_options["cong_algo"] = f"-K {self.options['cong_algo']}"

        elif netperf_options["testname"] == "-t UDP_STREAM":
            test_options = copy.copy(NetperfRunner.netperf_udp_options)

        netperf_options_list = list(netperf_options.values())
        netperf_options_string = " ".join(netperf_options_list)
        test_options_list = list(test_options.values())
        test_options_string = " ".join(test_options_list)

        if self.start_time > 0:
            sleep(self.start_time)

        super().run(
            partial(
                run_netperf,
                self.ns_id,
                netperf_options_string,
                self.destination_address.get_addr(with_subnet=False),
                test_options_string,
                self.destination_address.is_ipv6(),
            ),
            error_string_prefix="Running netperf",
        )

    @handle_keyboard_interrupt
    def parse(self):
        """
        Parse netperf output from `self.out`
        """
        self.out.seek(0)  # rewind to start of the temp file
        raw_stats = self.out.read().decode()

        # pattern that matches the netperf output corresponding to throughput
        throughput_pattern = r"NETPERF_INTERIM_RESULT\[\d+]=(?P<throughput>\d+\.\d+)"
        throughputs = [
            throughput.group("throughput")
            for throughput in re.finditer(throughput_pattern, raw_stats)
        ]

        # pattern that matches the netperf output corresponding to interval
        timestamp_pattern = r"NETPERF_ENDING\[\d+]=(?P<timestamp>\d+\.\d+)"
        timestamps = [
            timestamp.group("timestamp")
            for timestamp in re.finditer(timestamp_pattern, raw_stats)
        ]

        # pattern that gives the remote port
        remote_port_pattern = r"remote port is (?P<remote>\d+)"
        remote_port = re.search(remote_port_pattern, raw_stats).group("remote")

        # List storing collected stats
        # First item as "meta" item with user given information
        stats_list = [self.get_meta_item()]

        # Trim last result, since netperf typically gives unrealisticly high throughput
        # towards the end
        for i in range(len(throughputs) - 1):
            stats_list.append(
                {
                    "timestamp": timestamps[i],
                    # Netperf provides throughput as sending rate from sender's side
                    "sending_rate": throughputs[i],
                }
            )
        destination_ip = self.destination_address.get_addr(with_subnet=False)
        stats_dict = {f"{destination_ip}:{remote_port}": stats_list}

        NetperfResults.add_result(self.ns_id, stats_dict)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""Class to handles Ldp related functionalities"""

from functools import partial
from nest.engine.dynamic_routing import run_ldpd
from nest.routing.route_daemons import RoutingDaemonBase


class Ldp(RoutingDaemonBase):
    """
    Handles Ldp related functionalities for frr.
    """

    def __init__(self, router_ns_id, interfaces, conf_dir, **kwargs):
        super().__init__(router_ns_id, interfaces, "ldpd", conf_dir, **kwargs)

    def create_basic_config(self):
        """
        Creates a file with basic configuration for ldp.
        Use base `add_to_config` directly for more complex configurations
        """
        self.add_to_config("mpls ldp")
        self.add_to_config(
            f" router-id {self.interfaces[0].address.get_addr(with_subnet=False)}"
        )
        self.add_to_config(" address-family ipv4")
        self.add_to_config(
            f"discovery transport-address {self.interfaces[0].address.get_addr(with_subnet=False)}"
        )

        for interface in self.interfaces:
            self.add_to_config(f"  interface {interface.id}")

        if self.log_file is not None:
            self.add_to_config(f"log file {self.log_file}")

        self.create_config()

    def run(self):
        """
        Runs the ldpd command
        """
        super().run(
            engine_func=partial(
                run_ldpd, self.router_ns_id, self.conf_file, self.pid_file
            )
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

""" Netperf commands """
from .exec import exec_exp_commands

# pylint: disable=too-many-arguments
def run_netperf(ns_id, netperf_options, destination_ip, test_options, ipv6, out, err):
    """
    Run netperf

    Parameters
    ----------
    ns_id : str
        network namespace to run netperf from
    netperf_options : str
        default options to run netperf command with
    destination_ip : str
        ip address of the destination namespace
    run_time : num
        total time to run netperf for
    test_options : str
        experiment related netperf options
    ipv6 : bool
        determines if destination_ip is ipv4/ipv6
    out : File
        temporary file to hold the stats
    err : File
        temporary file to hold any errors

    Returns
    -------
    int
        return code of the command executed
    """
    if ipv6:
        return exec_exp_commands(
            f"ip netns exec {ns_id} netperf -6 {netperf_options} "
            f"-H {destination_ip} -- {test_options}",
            stdout=out,
            stderr=err,
        )

    return exec_exp_commands(
        f"ip netns exec {ns_id} netperf -4 {netperf_options} -H {destination_ip} -- {test_options}",
        stdout=out,
        stderr=err,
    )


def run_netserver(ns_id):
    """
    Run netserver

    Parameters
    ----------
    ns_id : str
        network namespace to run netperf from

    Returns
    -------
    int
        return code of the command executed
    """
    return exec_exp_commands(f"ip netns exec {ns_id} netserver")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Test Address sub-module in topology"""

import unittest
from nest.topology.address import Address, Subnet

# pylint: disable=missing-docstring


class TestTopologyAddress(unittest.TestCase):
    def setUp(self):
        self.addr = "10.0.0.1/24"
        self.addr_without_subnet = "10.0.0.1"
        self.default_addr = "DEFAULT"
        self.subnet_addr = "10.0.0.0/24"

    def test_get_addr(self):
        address = Address(self.addr)
        self.assertEqual(address.get_addr(), self.addr)
        self.assertEqual(address.get_addr(with_subnet=False), self.addr_without_subnet)

        with self.assertRaises(ValueError):
            Address("10.0.0.0a/24")  # invalid character 'a'

        with self.assertRaises(ValueError):
            Address("10.0asdfasf")  # invalid string

        with self.assertRaises(ValueError):
            Address("10.0.0./24")  # invalid ./

        with self.assertRaises(ValueError):
            Address("10.0.0.2000/24")  # invalid mask 2000

        with self.assertRaises(ValueError):
            Address("256.0.0.0/24")  # invalid mask 256

    def test_address_subnet(self):
        subnet_address = Address(self.subnet_addr)
        self.assertTrue(subnet_address.is_subnet())

        address = Address(self.addr)
        self.assertEqual(address.get_subnet(), self.subnet_addr)

        with self.assertRaises(ValueError):
            Subnet("10.0.0.1/24")  # 10.0.0.1/24 is not a subnet

    def test_subnet(self):
        subnet = Subnet(self.subnet_addr)
        self.assertEqual(subnet.get_next_addr().get_addr(), "10.0.0.1/24")
        self.assertEqual(subnet.get_next_addr().get_addr(), "10.0.0.2/24")
        self.assertEqual(subnet.get_next_addr().get_addr(), "10.0.0.3/24")

        with self.assertRaises(ValueError):
            Subnet("10.0.0.1/24").get_next_addr()

        with self.assertRaises(ValueError):
            Subnet(
                "10.0.0.0/32"
            ).get_next_addr()  # no next address possible as the mask is 32

    def test_default(self):
        def_addr = Address("DEFAULT")
        self.assertEqual("default", def_addr.get_addr())
        self.assertEqual(False, def_addr.is_subnet())

        with self.assertRaises(Exception):
            def_addr.get_subnet()  # no subnet for default

        with self.assertRaises(Exception):
            Subnet("DEFAULT")  # DEFAULT cannot be a subnet


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Traffic control management"""

from .exec import exec_subprocess

# Only bandwidth and latency is considered
# Assuming tc on egress
# Using Netem


def add_traffic_control(host_name, dev_name, rate, latency):
    """
    Add traffic control to host

    Parameters
    ----------
    host_name : str
        name of the host namespace
    rate : str
        rate of the bandwidth
    latency : str
        latency of the link
    dev_name : str
    """
    exec_subprocess(
        f"tc -n {host_name} qdisc add dev {dev_name} root"
        f" netem rate {rate} latency {latency}"
    )


def add_qdisc(ns_name, dev_name, qdisc, parent="", handle="", **kwargs):
    """
    Add a qdisc on an device

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the device
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the qdisc in major:0 form (Default value = '')
    dev_name : str
        name of the device
    """
    if parent and parent != "root":
        parent = "parent " + parent

    if handle:
        handle = "handle " + handle

    qdisc_params = ""
    for param, value in kwargs.items():
        qdisc_params += param + " " + value + " "

    exec_subprocess(
        f"tc -n {ns_name} qdisc add dev {dev_name}"
        f" {parent} {handle} {qdisc} {qdisc_params}"
    )


def change_qdisc(ns_name, dev_name, qdisc, parent, handle, **kwargs):
    """
    Change a qdisc that is already present on an device

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the device
    parent : str
        id of the parent class in major:minor form(optional)
    handle : str
        id of the qdisc in major:0 form
    dev_name : str
        name of the device
    """
    if parent and parent != "root":
        parent = "parent " + parent

    if handle:
        handle = "handle " + handle

    qdisc_params = ""
    for param, value in kwargs.items():
        qdisc_params += param + " " + value + " "

    exec_subprocess(
        f"tc -n {ns_name} qdisc change dev {dev_name}"
        f" {parent} {handle} {qdisc} {qdisc_params}"
    )


def replace_qdisc(ns_name, dev_name, qdisc, parent="", handle="", **kwargs):
    """
    Replace a qdisc that is already present on an device

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the device
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the qdisc in major:0 form (Default value = '')
    dev_name : str
        name of the device
    """
    if parent and parent != "root":
        parent = "parent " + parent

    if handle:
        handle = "handle " + handle

    qdisc_params = ""
    for param, value in kwargs.items():
        qdisc_params += param + " " + value + " "

    exec_subprocess(
        f"tc -n {ns_name} qdisc replace dev {dev_name}"
        f" {parent} {handle} {qdisc} {qdisc_params}"
    )


def delete_qdisc(ns_name, dev_name, parent, handle):
    """
    Delete a qdisc on an device

    Parameters
    ----------
    ns_name : str
        name of the namespace
    qdisc : str
        qdisc used on the device
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the qdisc in major:0 form (Default value = '')
    dev_name : str
        name of the device
    """
    if parent and parent != "root":
        parent = "parent " + parent

    if handle:
        handle = "handle " + handle

    exec_subprocess(f"tc -n {ns_name} qdisc del dev {dev_name} {parent} {handle}")


def add_class(ns_name, dev_name, parent, qdisc, classid="", **kwargs):
    """
    Add a class to a qdisc

    Parameters
    ----------
    ns_name : str
        name of the namespace
    parent : str
        id of the parent class in major:minor form(optional)
    qdisc : str
        qdisc used on the device
    classid : str
        id of the class in major:minor form (Default value = '')
    dev_name : str
        name of the device
    """
    if classid:
        classid = "classid " + classid

    qdisc_params = ""
    for param, value in kwargs.items():
        qdisc_params += param + " " + value + " "

    exec_subprocess(
        f"tc -n {ns_name} class add dev {dev_name} parent {parent}"
        f" {classid} {qdisc} {qdisc_params}"
    )


def change_class(ns_name, dev_name, parent, qdisc, classid, **kwargs):
    """
    Change a class that is already present on an device

    Parameters
    ----------
    ns_name : str
        name of the namespace
    parent : str
        id of the parent class in major:minor form(optional)
    qdisc : str
        qdisc used on the device
    classid : str
        id of the class in major:minor form (Default value = '')
    dev_name : str
        name of the device
    """

    if classid:
        classid = "classid " + classid

    qdisc_params = ""
    for param, value in kwargs.items():
        qdisc_params += param + " " + value + " "

    exec_subprocess(
        f"tc -n {ns_name} class change dev {dev_name} parent {parent}"
        f" {classid} {qdisc} {qdisc_params}"
    )


def delete_class(ns_name, dev_name, parent, classid):
    """
    Delete a class added to a qdisc
    Parameters
    ----------
    ns_name : str
        name of the namespace
    dev_name : str
        name of the device
    parent : str
        id of the parent class in major:minor form(optional)
    classid : str
        id of the class in major:minor form (Default value = '')
    """

    if classid:
        classid = "classid " + classid

    exec_subprocess(
        f"tc -n {ns_name} class delete dev {dev_name} parent {parent}" f" {classid}"
    )


# pylint: disable=too-many-arguments
def add_filter(
    ns_name, dev_name, protocol, priority, filtertype, parent="", handle="", **kwargs
):
    """
    Add a filter to a class

    Parameters
    ----------
    ns_name : str
        name of the namespace
    protocol : str
        protocol used
    priority : str
        priority
    filtertype : str
        one of the available filters
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the filter (Default value = '')
    qdisc : str
        qdisc used on the device
    dev_name : str
        name of the device
    """

    # TODO: Check if protocol can be removed from the arguments since it's always IP

    if parent and parent != "root":
        parent = "parent " + parent

    if handle:
        handle = "handle " + handle

    filter_params = ""

    for param, value in kwargs.items():
        filter_params += param + " " + value + " "

    exec_subprocess(
        f"tc -n {ns_name} filter add dev {dev_name} {parent} {handle}"
        f" protocol {protocol} prio {priority} {filtertype} {filter_params}"
    )


def delete_filter(ns_name, dev_name, parent, handle):
    """
    Delete filter on a class

    Parameters
    ----------
    ns_name : str
        name of the namespace
    dev_name : str
        name of the device
    parent : str
        id of the parent class in major:minor form(optional) (Default value = '')
    handle : str
        id of the filter (Default value = '')
    """
    if parent and parent != "root":
        parent = "parent " + parent

    if handle:
        handle = "handle " + handle

    exec_subprocess(f"tc -n {ns_name} filter delete dev {dev_name} {parent} {handle}")


def get_tc_version():
    """
    check for current tc version

    Returns
    -------
    str
        string containing current tc version
    """
    return exec_subprocess("tc -V", output=True)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Network Namespace management"""

from .exec import exec_subprocess


def create_ns(ns_name):
    """
    Create namespace if it doesn't already exist

    Parameters
    ----------
    ns_name : str
        namespace name
    """
    exec_subprocess(f"ip netns add {ns_name}")


def delete_ns(ns_name):
    """
    Drops the namespace if it already exists.

    Parameters
    ----------
    ns_name : str
        namespace name
    """
    exec_subprocess(f"ip netns del {ns_name}")


def kill_all_processes(ns_name):
    """
    Kill all processes in a namespace

    Parameters
    ----------
    ns_name : str
        Namespace name
    """

    exec_subprocess(f"kill $(ip netns pids {ns_name})", shell=True)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to creation of IFB for qdiscs in topology"""

import logging
import nest.config as config
from nest import engine
from nest.topology.device import Device

logger = logging.getLogger(__name__)

# pylint: disable=too-many-instance-attributes


class Ifb(Device):
    """
    This is the IFB created to be able to add two classless
    qdiscs, netem and another on a single interface

    Attributes
    ----------
    id : str
        This value is used by `engine` to create emulated IFB
        entity
    node_id : str
        id of the Node to  which this IFB belongs
    veth_end_id : str
        The id of the interface that the IFB is attached to
    bandwidth : str
        The bandwidth set to the device
    qdisc : str
        The qdisc assigned to the interface
    """

    def __init__(self, name, node_id, veth_end_id):
        """
        Constructor for an IFB

        Created only if bandwidth is given or default bandwidth is set
        in case of adding qdisc and delay buy bandwidth not specified

        Parameters
        ----------
        name : str
            User given name for the interface
        node_id : str
            This is the id of the node that the device belongs to
        veth_end_id : str
            This is the id of the veth that the IFB is attached to
        """

        super().__init__(name, node_id)
        self.veth_end_id = veth_end_id

        # Create Ifb and add to namespace
        engine.create_ifb(self._id)
        engine.add_int_to_ns(node_id, self._id)

        self.set_mode("UP")

        self.current_bandwidth = config.get_value("default_bandwidth")
        self._add_default_qdisc_and_mirror_packets()

    def _add_default_qdisc_and_mirror_packets(self):
        """
        Sets default bandwidth to the IFB
        """

        default_route = {"default": "1"}

        # TODO: find how to set a good bandwidth
        default_bandwidth = {"rate": config.get_value("default_bandwidth")}

        # TODO: Standardize this; seems like arbitrary handle values
        # were chosen.
        # HTB class is added since, to use a filter and redirect traffic,
        # a classid is needed and htb gives it that, since it's a class
        self.add_qdisc("htb", "root", "1:", **default_route)
        self.add_class("htb", "1:", "1:1", **default_bandwidth)
        self.add_qdisc("pfifo", "1:1", "11:")

        action_redirect = {
            "match": "u32 0 0",  # from man page examples
            "action": "mirred",
            "egress": "redirect",
            "dev": self.id,
        }

        # NOTE: Use Filter API
        # Action mirred, redicting traffic, etc is needed since netem and
        # the user giver qdisc are both classless and cannot be added to
        # the same device
        engine.add_filter(
            self.node_id,
            self.veth_end_id,
            "all",
            "1",
            "u32",
            parent="1:",
            **action_redirect,
        )

    def set_bandwidth(self, bandwidth):
        """
        Set a maximum bandwidth for the ifb

        Paramaters
        ----------
        bandwidth : str
            The bandwidth that has to be set with the units
        """

        self.current_bandwidth = bandwidth

        bandwidth_parameter = {"rate": bandwidth}
        self.change_class("htb", "1:", "1:1", **bandwidth_parameter)

    def set_qdisc(self, qdisc, **kwargs):
        """
        Adds the Queueing algorithm to the interface

        Parameters
        ----------
        qdisc : string
            The Queueing algorithm to be added
        bandwidth :
            Link bandwidth
        """

        current_bandwidth_parameter = {"rate": self.current_bandwidth}

        engine.change_class(
            self.node_id, self.id, "1:", "htb", "1:1", **current_bandwidth_parameter
        )

        self.delete_qdisc("11:")
        self.add_qdisc(qdisc, "1:1", "11:", **kwargs)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
This module handles the validation of inputs passed to methods
(mainly for client-side APIs).

All methods which require input validation should be decorated by
`@input_validator`.

For eg., the below method called `samplemethod` is decorated by
`@input_validator`:

    @input_validator
    def samplemethod(arg1: Type1, arg2: Type2):
        ...

When the above method is invoked:

    samplemethod(input1, input2)

The decorator `@input_validator` validates if `input1` is of type
`Type1` and if `input2` is of type `Type2`. Say, `input1` is not
of `Type1`. Then, it will additionally check if `input1` can be
safely typecasted to `Type1`. If even this is not possible, then
the decorator throws an exception.

For figuring out if `input1` can be typecasted to `Type1`, there are
two requirements for the class `Type1`:
  * The class is expected to have a constructor with 1 parameter
    (excluding self). The constructor should raise ValueError if there
    is any error in validation of parameters.
  * The class should have a static method called `allowed_type_cast()`.
    It should return a list of types that can safely be typecasted to
    `Type1`. The logic for typecasting should be handled in the constructor.

If the above two requirements are not met, then `@input_validator` decorator
will not attempt to typecast the input.

NOTE: If a method has more than one decorator, then input_validator should
be applied before all other decorators.

    @some_other_decorator
    @input_validator
    def samplemethod(arg1: Type1, arg2: Type2):
        ...
"""

from .input_validator import input_validator



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Class to handles RIP related functionalities"""

from functools import partial
from nest.engine.dynamic_routing import run_ripd
from nest.routing.route_daemons import RoutingDaemonBase


class Rip(RoutingDaemonBase):
    """
    Handles RIP related functionalities.
    Refer to `DaemonBase` for usage
    """

    def __init__(self, router_ns_id, interfaces, conf_dir, **kwargs):
        super().__init__(router_ns_id, interfaces, "ripd", conf_dir, **kwargs)

    def add_rip(self):
        """
        Add command to enable RIP on router to config file
        """
        if self.ipv6:
            self.add_to_config("router ripng")
        else:
            self.add_to_config("router rip")

    def add_network(self, network):
        """
        Add command for subnet or interface to run RIP on to config file

        Parameters
        ----------
        network : str
            subnet or interface to run RIP on
        """
        self.add_to_config(f" network {network}")

    def create_basic_config(self):
        """
        Creates a file with basic configuration for RIP.
        Use base `add_to_config` directly for more complex configurations
        """
        self.add_rip()
        for interface in self.interfaces:
            self.add_network(interface.id)
        if self.log_file is not None:
            self.add_to_config(f"log file {self.log_file}")

    def run(self):
        """
        Runs the ripd command
        """
        super().run(
            engine_func=partial(
                run_ripd, self.router_ns_id, self.conf_file, self.pid_file, self.ipv6
            )
        )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Helper function for logging"""

import logging
import re

# See: https://stackoverflow.com/a/35804945
def add_logging_level(level_name, level_num, method_name=None):
    """
    Comprehensively adds a new logging level to the `logging` module and the
    currently configured logging class.

    `levelName` becomes an attribute of the `logging` module with the value
    `levelNum`. `methodName` becomes a convenience method for both `logging`
    itself and the class returned by `logging.getLoggerClass()` (usually just
    `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
    used.

    To avoid accidental clobberings of existing attributes, this method will
    raise an `AttributeError` if the level name is already an attribute of the
    `logging` module or if the method name is already present

    Example
    -------
    >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
    >>> logging.getLogger(__name__).setLevel("TRACE")
    >>> logging.getLogger(__name__).trace('that worked')
    >>> logging.trace('so did this')
    >>> logging.TRACE
    5

    """
    if not method_name:
        method_name = level_name.lower()

    if hasattr(logging, level_name):
        raise AttributeError("{} already defined in logging module".format(level_name))
    if hasattr(logging, method_name):
        raise AttributeError("{} already defined in logging module".format(method_name))
    if hasattr(logging.getLoggerClass(), method_name):
        raise AttributeError("{} already defined in logger class".format(method_name))

    def log_for_level(self, message, *args, **kwargs):
        if self.isEnabledFor(level_num):
            # pylint: disable=protected-access
            self._log(level_num, message, args, **kwargs)

    def log_to_root(message, *args, **kwargs):
        logging.log(level_num, message, *args, **kwargs)

    logging.addLevelName(level_num, level_name)
    setattr(logging, level_name, level_num)
    setattr(logging.getLoggerClass(), method_name, log_for_level)
    setattr(logging, method_name, log_to_root)


def get_trace_filehandler():
    """
    Creates and returns a file handler for logging engine commands
    """
    filehandler = logging.FileHandler("commands.sh", "w")
    # pylint: disable=no-member
    filehandler.setLevel(logging.TRACE)
    formatter = logging.Formatter("%(message)s")
    filehandler.setFormatter(formatter)
    filehandler.addFilter(lambda record: record.levelno == logging.TRACE)

    return filehandler


def update_nest_logger(level):
    """
    Update top logger when log level is updated via config

    Parameters
    ----------
    level: str
        log level from config
    """
    nest_logger = logging.getLogger(
        __name__.split(".")[0]
    )  # get the root's child logger
    nest_logger.setLevel(level)  # Update logger level
    # Update handler level
    nest_handler = nest_logger.handlers[0]
    nest_handler.setLevel(level)
    # pylint: disable=no-member
    if level == "TRACE" and not any(
        handler.level == logging.TRACE and handler is logging.FileHandler
        for handler in nest_logger.handlers
    ):  # Avoid adding multiple trace filehandlers
        nest_logger.addHandler(get_trace_filehandler())


# pylint: disable=too-few-public-methods
class DuplicateFilter(logging.Filter):
    """
    Super class for duplicate filter
    """

    def __init__(self):
        super().__init__()
        self.messages = set()
        self.filter_pat = r".*"

    def filter(self, record):
        """
        Checks whether log message is already
        logged and adds it to the current set of
        logged message accordingly
        """
        log_message = record.getMessage()

        if log_message in self.messages:
            return 0

        if re.match(self.filter_pat, log_message) is not None:
            self.messages.add(log_message)
        return 1


# pylint: disable=too-few-public-methods
class DepedencyCheckFilter(logging.Filter):
    """
    Filters duplicate depedency missing logs.
    """

    def __init__(self):
        super().__init__()
        self.messages = set()

        # Assuming the depedency error is of the form "{depedency} not found".
        self.filter_pat = r"(\w|\d|-|_)+ not found\."


class DuplicateRoutingLogsFilter(DuplicateFilter):
    """
    Filters duplicate "quagga/frr logging enabled" logs
    """

    def __init__(self):
        super().__init__()
        self.messages = set()

        # Assuming the log is of the form
        # "{frr/quagga} logging enabled. Log files can found in {dir} directory".
        # pylint: disable=line-too-long
        self.filter_pat = r"(quagga|frr) logging enabled\. Log files can found in (\w|\d|-|_|:)+ directory"



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""Plot ping results"""

import logging
import matplotlib.pyplot as plt
from nest.experiment.interrupts import handle_keyboard_interrupt
from ..pack import Pack
from .common import simple_plot

logger = logging.getLogger(__name__)


def _plot_ping_flow(flow, node, dest):
    """
    Plot ping stats of the flow

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    flow : List
        List with timestamps and stats
    node : str
        Node from which ping results were obtained from
    dest :
        Destination ip:port address of the flow

    Returns
    -------
    tuple/None
        Timestamped rtt values
    """
    # "meta" item will always be present, hence `<= 1`
    if len(flow) <= 1:
        logger.warning(
            "Flow from %s to destination %s " "doesn't have any parsed ping result.",
            node,
            dest,
        )
        return None

    # First item is the "meta" item with user given information
    user_given_start_time = float(flow[0]["start_time"])
    destination_node = flow[0]["destination_node"]

    # "Bias" actual start_time in experiment with user given start time
    start_time = float(flow[1]["timestamp"]) - user_given_start_time

    timestamp = []
    rtt = []

    for data in flow[1:]:
        rtt.append(float(data["rtt"]))
        relative_time = float(data["timestamp"]) - start_time
        timestamp.append(relative_time)

    fig = simple_plot(
        "",
        timestamp,
        rtt,
        "Time (Seconds)",
        "Ping Latency (ms)",
        legend_string=f"{node} to {destination_node} ({dest})",
    )
    filename = f"ping_{node}_to_{destination_node}({dest}).png"
    Pack.dump_plot("ping", filename, fig)
    plt.close(fig)

    return (timestamp, rtt)


@handle_keyboard_interrupt
def plot_ping(parsed_data):
    """
    Plot statistics obtained from ping

    Parameters
    ----------
    exp_name : str
        Name of experiment for which results were obtained
    parsed_data : Dict
        JSON data parsed from ss
    """
    for node in parsed_data:
        node_data = parsed_data[node]

        for connection in node_data:
            for dest in connection:
                flow = connection[dest]
                _plot_ping_flow(flow, node, dest)



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal

"""API related to creation of virtual ethernet devices in topology"""

import logging
import nest.config as config
import nest.global_variables as g_var
from nest import engine
from .device import Device
from .address import Address

logger = logging.getLogger(__name__)

# pylint: disable=too-many-instance-attributes


class VethEnd(Device):
    """
    This is the end of the veth created, to connect two devices

    Attributes
    ----------
    id : str
        This value is used by `engine` to create emulated interface
        entity
    node_id : str
        id of the Node to  which this Interface belongs
    address : str/Address
        IP address assigned to this interface
    bandwidth : str
        The bandwidth set to the device
    delay : str
        The delay set to the device
    """

    def __init__(self, name, node_id):
        """
        Constructor of VethEnd

        Parameters
        ----------
        name : str
            Name of the interface
        node_id : str
            Id of the node that the interface belongs to
        """

        super().__init__(name, node_id)
        self._address = None
        self.bandwidth = None
        self.delay = None
        self._set_structure = False

        self._mtu = 1500
        self._is_mpls_enabled = False

    @property
    def subnet(self):
        """Getter for the subnet to which the address belongs to"""
        return self.address.get_subnet()

    @property
    def address(self):
        """
        Getter for the address associated
        with the interface
        """
        return self._address

    @address.setter
    def address(self, address):
        """
        Assigns IP address to an interface

        Parameters
        ----------
        address : Address or str
            IP address to be assigned to the interface
        """
        # TODO: The below check is redundant since
        # input_validator check is run in Interface
        if isinstance(address, str):
            address = Address(address)

        if self.node_id is not None:
            engine.assign_ip(self.node_id, self.id, address.get_addr())
            self._address = address
        else:
            # TODO: Create our own error class
            raise NotImplementedError(
                "You should assign the interface to node or router before assigning address to it."
            )

        # Global variable to check if address is ipv6 or not for DAD check
        if address.is_ipv6() is True:
            g_var.IS_IPV6 = True

    def disable_ip_dad(self):
        """
        Disables Duplicate addresses Detection (DAD) for an interface.
        """
        engine.disable_dad(self.node_id, self._id)

    def enable_mpls(self):
        """
        Enables mpls input through the interface.
        Requires mpls kernel modules to be loaded.

        Run ``sudo modprobe mpls_iptunnel`` to load mpls modules.
        """
        if self.node_id is None:
            raise NotImplementedError(
                "You should assign the interface to node or router before enabling mpls"
            )

        engine.set_mpls_max_label_node(self.node_id, int(100000))

        if self._is_mpls_enabled is False:
            engine.enable_mpls_interface(self.node_id, self.id)
            self._is_mpls_enabled = True
            self.mtu = 1504

    def is_mpls_enabled(self):
        """
        Check if the interface has mpls enabled
        """
        return self._is_mpls_enabled

    @property
    def mtu(self):
        """
        Get the maximum transmit unit value for the interface
        """
        return self._mtu

    @mtu.setter
    def mtu(self, mtu_value):
        """
        Set the maximum transmit unit value for the interface
        Default is 1500 bytes.
        """
        if self._mtu != mtu_value:
            engine.set_mtu_interface(self.node_id, self.id, int(mtu_value))
            self._mtu = mtu_value
            logger.debug("MTU of interface %s set to %s", self.name, str(self.mtu))

    def set_structure(self):
        """
        Sets a proper structure to the interface by creating HTB class
        with default bandwidth and a netem qdisc as a child.
        HTB is used for setting bandwidth and netem is used for setting
        delay, packet corruption etc
        (default bandwidth = 1024mbit)
        """

        if self._set_structure is True:
            return

        self._set_structure = True

        default_route = {"default": "1"}

        # HTB is added since netem is a classless qdisc. So, htb class,
        # With netem as child is added
        self.add_qdisc("htb", "root", "1:", **default_route)

        # TODO: find how to set a good bandwidth
        default_bandwidth = {"rate": config.get_value("default_bandwidth")}

        self.add_class("htb", "1:", "1:1", **default_bandwidth)

        self.add_qdisc("netem", "1:1", "11:")

    def enable_offload(self, offload_name):
        """
        API for enabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to enable
        """
        if not isinstance(offload_name, list):
            offload_name = [offload_name]
        valid_offloads(offload_name)
        namespace_id = self.node_id
        interface_id = self.id
        for offload_type in offload_name:
            if engine.ethtool.enable_offloads(namespace_id, interface_id, offload_type):
                logger.debug(
                    "%s is enabled on interface %s",
                    offload_type,
                    self.name,
                )
            else:
                logger.error(
                    "%s is not enabled on interface %s",
                    offload_type,
                    self.name,
                )

    def disable_offload(self, offload_name):
        """
        API for disabling offloads
        Parameters
        ----------
        offload_name : str
            The type of offload names that need to disable
        """
        if not isinstance(offload_name, list):
            offload_name = [offload_name]
        valid_offloads(offload_name)
        namespace_id = self.node_id
        interface_id = self.id
        for offload_type in offload_name:
            if engine.ethtool.disable_offloads(
                namespace_id, interface_id, offload_type
            ):
                logger.debug(
                    "%s is disabled on interface %s",
                    offload_type,
                    self.name,
                )
            else:
                logger.error(
                    "%s is not disabled on interface %s",
                    offload_type,
                    self.name,
                )


def valid_offloads(offload_name):
    """
    Check valid offloads

    Parameters
    -----------
    offload_name : str
        The offload name
    """
    offloads_list = ["tso", "gso", "gro"]
    for offload_type in offload_name:
        if not offload_type in offloads_list:
            logger.error("Invalid offload")
            raise ValueError(f"{offload_type} is not a valid offload")



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2021 NITK Surathkal
"""enabling/disabling offloads"""
import logging
from .exec import exec_subprocess

logger = logging.getLogger(__name__)


def enable_offloads(namespace_id, interface_id, offload_type):
    """
    enable offload
    Parameters
    ----------
    namespace_id : str
        The namespace id on which interface is available
    interface_id : str
            The interface id where offloads enable
    offload_type : str
            The type of offloads that need to be enabled

    Returns
    --------
    bool
        success of ethtool command
    """
    status = exec_subprocess(
        f"ip netns exec {namespace_id} ethtool -K {interface_id} {offload_type} on "
    )
    return status == 0


# Disable Offloads
def disable_offloads(namespace_id, interface_id, offload_type):
    """
    disable offload
    Parameters
    ----------
    namespace_id : str
        The namespace id on which interface is available
    interface_id : str
        The interface id where offloads disable
    offload_type : str
        The type of offloads that need to be disabled

    Returns
    -------
    bool
        success of ethtool command
    """
    status = exec_subprocess(
        f"ip netns exec {namespace_id} ethtool -K {interface_id} {offload_type} off "
    )
    return status == 0



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2020 NITK Surathkal

"""IP route commands"""

from .exec import exec_subprocess


def add_route(host_name, dest_ip, next_hop_ip, via_int):
    """
    Adds a route in routing table of host.

    Parameters
    ----------
    host_name : str
        name of the host namespace
    dest_ip : str
        the destination ip for the route
    next_hop_ip : str
        IP of the very next interface
    via_int : str
        the corresponding interface in the host
    """
    exec_subprocess(
        f"ip netns exec {host_name} ip route add {dest_ip}"
        f" via {next_hop_ip} dev {via_int}"
    )



# SPDX-License-Identifier: GPL-2.0-only
# Copyright (c) 2019-2022 NITK Surathkal

"""
Helper classes and methods for client-side input validation.
"""

import typing
import functools
import logging
from builtins import isinstance
from inspect import getfullargspec
from .typing_helper_methods import (
    get_inner_type_of_optional_field,
    is_list,
    is_optional,
)

logger = logging.getLogger(__name__)


def input_validator(func):
    """
    Decorator to validate the input provided by the user for function `func`.
    This also typecasts the parameters of `func` as per the provided
    type hints in the function.

    Each parameter in `func` is expected to have a type hint. Type hints
    are necessary for input validation. If type hint is not provided for a
    parameter, then the input for the parameter will not be validated and
    type casted.

    Note: Default values of parameters are not validated.

    Parameters
    ----------
    func: Function
        Decorated function whose types are validated

    Returns
    -------
    func(*args, **kwargs)
        If all args are valid, then calls the decorated function with
        typecasted parameters.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        argspec = getfullargspec(func)

        # Forward references type hints:
        # https://www.python.org/dev/peps/pep-0563/#resolving-type-hints-at-runtime
        annotations = typing.get_type_hints(func)

        casted_args = []
        casted_kwargs = {}

        for (arg_name, arg_value) in zip(argspec.args, args):
            casted_arg_value = arg_value
            if arg_name in annotations:
                casted_arg_value = validate_input_and_cast(
                    func.__qualname__, arg_name, arg_value, annotations[arg_name]
                )
            elif arg_name != "self":
                logger.debug(
                    "Parameter %s of function %s does not have a type hint",
                    arg_name,
                    func.__qualname__,
                )
            casted_args.append(casted_arg_value)

        # Default values are not checked here, it is assumed that correct
        # values are passed for default parameters
        for kwarg_name in kwargs:
            kwarg_value = kwargs[kwarg_name]
            casted_kwarg_value = kwarg_value
            if kwarg_name in annotations:
                casted_kwarg_value = validate_input_and_cast(
                    func.__qualname__, kwarg_name, kwarg_value, annotations[kwarg_name]
                )
            else:
                logger.debug(
                    "Parameter %s of function %s does not have a type hint",
                    kwarg_name,
                    func.__qualname__,
                )
            casted_kwargs[kwarg_name] = casted_kwarg_value

        return func(*casted_args, **casted_kwargs)

    return wrapper


def validate_input_and_cast(func_name, arg_name, arg_value, expected_type):
    """
    Validate input `arg_value` provided by user for parameter `arg_name`
    of function `func_name`. The type of `arg_value` is expected to be
    `expected_type`.

    Returns the `arg_value` if it is of expected_type. If the `arg_value`
    can be typecasted to `expected_type`, then the type casted value is
    returned. Else TypeError exception is raised.

    Parameters
    ----------
    func_name: string
        Name of the function whose parameters are being validated
    arg_name: string
        Name of the parameter being validated.
    arg_value: object
        The input value for the parameter given by the client/user.
    expected_type: type
        The expected type of `arg_value`. Forward referencing is supported:
        https://www.python.org/dev/peps/pep-0484/#forward-references.
        So, `expected_type` can be a string which evalautes to a type.

    Returns
    -------
    object
        Return true if `client__input` is of type `expected_type`.
        Else returns false.
    """
    if is_optional(expected_type):
        # For optional field, check if arg_value is None.
        if arg_value is None:
            return arg_value

        # Strip out the optional field tag, for compatibility with
        # older python versions (python 3.7)
        expected_type = get_inner_type_of_optional_field(expected_type)

    if is_list(expected_type):
        # For list, we only validate that `arg_value` is a list.
        # We do not validate the inner contents of the list.
        if isinstance(arg_value, list):
            return arg_value

        raise TypeError(
            f"Expected type of argument '{arg_name}' in method '{func_name}' is List. "
            f"But got input '{arg_value}' of type {type(arg_value)}"
        )

    if isinstance(arg_value, expected_type):
        return arg_value

    if hasattr(expected_type, "allowed_type_cast"):
        for casting_type in expected_type.allowed_type_cast():
            if isinstance(arg_value, casting_type):
                try:
                    return expected_type(arg_value)
                except ValueError as error:
                    raise TypeError(
                        f"For argument '{arg_name}' in method '{func_name}', "
                        f"converting '{arg_value}' to type {expected_type.__name__} failed.\n"
                        f"Please see the previous exception to know why the conversion failed."
                    ) from error

    raise TypeError(
        f"Expected type of argument '{arg_name}' in method '{func_name}' is {expected_type}. "
        f"But got input '{arg_value}' of type {type(arg_value)}"
    )
